from qxmt.configs import DeviceConfig
from qxmt.devices.base import BaseDevice


class DeviceBuilder:
    """
    Builder class for quantum devices.

    Examples:
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1000,
        ... )
        >>> device = DeviceBuilder(config).build()
    """

    def __init__(
        self,
        config: DeviceConfig,
    ) -> None:
        """Initialize the device builder.

        Args:
            config (DeviceConfig): Configuration for the quantum device. It is element of the ExperimentConfig.
        """
        self.config: DeviceConfig = config

    def build(self) -> BaseDevice:
        """Build a quantum device. it can be a general-purpose device overseeing multiple platforms.

        Returns:
            BaseDevice: General-purpose device overseeing multiple platforms
        """
        return BaseDevice(
            platform=self.config.platform,
            name=self.config.name,
            n_qubits=self.config.n_qubits,
            shots=self.config.shots,
            random_seed=self.config.random_seed,
        )



import multiprocessing as mp
import os
from pathlib import Path
from typing import Any

import pennylane as qml
import pytz

# set module path for developer environment
MODULE_HOME: Path = Path(__file__).resolve().parents[1]

# set current working directory for user environment
CURRENT_WORKING_DIR: Path = Path.cwd()

# set project root directory.
# it swith by environment variable "QXMT_ENV"
PROJECT_ROOT_DIR: Path = (
    MODULE_HOME if os.getenv("QXMT_ENV", "").lower() == "dev" else CURRENT_WORKING_DIR.resolve().parents[0]
)

# set default experiment directory.
# it swith by environment variable "QXMT_ENV"
DEFAULT_EXP_DIRC: Path = PROJECT_ROOT_DIR / "experiments"

# set default experiment file name.
DEFAULT_EXP_DB_FILE: Path = Path("experiment.json")

# set default config file name.
DEFAULT_EXP_CONFIG_FILE: Path = Path("config.yaml")

# set supported quantum platforms and devices
PENNYLANE_PLATFORM: str = "pennylane"
SUPPORTED_PLATFORMS: list[str] = [PENNYLANE_PLATFORM]
PENNYLANE_DEVICES: tuple[Any, ...] = (qml.devices.Device, qml.devices.LegacyDevice, qml.devices.QubitDevice)

# set default model name
DEFAULT_MODEL_NAME: str = "model.pkl"

# set default shot results file name
DEFAULT_SHOT_RESULTS_NAME: str = "shots.h5"

# set default LLM for generating description
LLM_MODEL_PATH = "microsoft/Phi-3-mini-128k-instruct"

# set default n_jobs for parallel processing
# it mainly used in kernel calculation, cross validation
DEFAULT_N_JOBS = max(1, int(mp.cpu_count() * 0.6))

# set default timezone
TZ: pytz.BaseTzInfo = pytz.timezone("Asia/Tokyo")

# set color mapt for visualization
DEFAULT_COLOR_MAP = "viridis"



import time
from typing import Any, Callable


def notify_long_running(func: Callable) -> Callable:
    """Decorator to notify the user that a function is running."""

    def wrapper(*args: dict, **kwargs: Any) -> Any:
        print(f'Executing "{func.__name__}". This may take some time...')
        result = func(*args, **kwargs)
        return result

    return wrapper


def retry_on_exception(retries: int, delay: float) -> Callable:
    """Decorator to retry a function if an exception is raised."""

    def decorator(func: Callable) -> Callable:
        def wrapper(*args: dict, **kwargs: Any) -> None:
            attempt = 0
            while attempt < retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    attempt += 1
                    if attempt >= retries:
                        raise e
                    time.sleep(delay)

        return wrapper

    return decorator



from pathlib import Path
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np

from qxmt.datasets.schema import Dataset


def plot_residual(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    grid: bool = True,
    save_path: Optional[str | Path] = None,
    **kwargs: Any,
) -> None:
    """Plot residuals vs predicted values.
    residuals caluculated by y_true - y_pred.

    Args:
        y_true (np.ndarray): actual values of the data.
        y_pred (np.ndarray): predicted values of the data.
        grid (bool, optional): grid on the plot. Defaults to True.
        save_path (Optional[str  |  Path], optional): save path of graph. Defaults to None.
        **kwargs (Any): additional arguments for plot.
            title (str, optional): title of the plot. Defaults to "Residual Plot".
    """
    assert len(y_true) == len(y_pred), "Length of y_true and y_pred must be the same."
    residuals = y_true - y_pred

    # Residuals vs Predicted values
    plt.figure(figsize=(7, 5), tight_layout=True)
    plt.scatter(y_pred, residuals, alpha=0.7, edgecolors="k", color="blue")
    plt.axhline(y=0, color="red", linestyle="--")

    plt.xlabel("Predicted Values")
    plt.ylabel("Residuals")
    plt.grid(grid)

    title = str(kwargs.get("title", '"Residual Plot"'))
    plt.title(title)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()


def plot_actual_vs_predicted(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    grid: bool = True,
    save_path: Optional[str | Path] = None,
    **kwargs: Any,
) -> None:
    """Plot actual vs predicted values.

    Args:
        y_true (np.ndarray): actual values of the data.
        y_pred (np.ndarray): predicted values of the data.
        grid (bool, optional): grid on the plot. Defaults to True.
        save_path (Optional[str  |  Path], optional): save path of graph. Defaults to None.
        **kwargs (Any): additional arguments for plot.
            title (str, optional): title of the plot. Defaults to "Residual Plot".
    """
    assert len(y_true) == len(y_pred), "Length of y_true and y_pred must be the same."

    plt.figure(figsize=(7, 5), tight_layout=True)

    # Scatter plot of actual vs predicted values
    plt.scatter(y_true, y_pred, alpha=0.7, edgecolors="k", color="blue")
    # Line of perfect predictions
    plt.plot([min(y_true), max(y_true)], [min(y_true), max(y_true)], color="red", linestyle="--", lw=2)

    plt.xlabel("Actual Values")
    plt.ylabel("Predicted Values")
    plt.grid(grid)

    title = str(kwargs.get("title", '"Actual vs Predicted Plot"'))
    plt.title(title)

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()



from qxmt.datasets.generate.linear import (
    generate_linear_regression_data,
    generate_linear_separable_data,
)
from qxmt.datasets.generate.loader import GeneratedDataLoader

__all__ = [
    "generate_linear_separable_data",
    "generate_linear_regression_data",
    "GeneratedDataLoader",
]



import inspect
from logging import Logger
from typing import Callable, Optional, cast, get_type_hints

import numpy as np
from sklearn.model_selection import train_test_split

from qxmt.configs import (
    DatasetConfig,
    ExperimentConfig,
    FileConfig,
    GenerateDataConfig,
    OpenMLConfig,
)
from qxmt.datasets.file import FileDataLoader
from qxmt.datasets.generate import GeneratedDataLoader
from qxmt.datasets.openml import OpenMLDataLoader
from qxmt.datasets.schema import Dataset
from qxmt.logger import set_default_logger
from qxmt.types import PROCESSCED_DATASET_TYPE, RAW_DATASET_TYPE
from qxmt.utils import load_object_from_yaml

LOGGER = set_default_logger(__name__)


class DatasetBuilder:
    """
    [NOTE]: Currently, this class is only support numpy array data type.
    The other data types will be supported in the future.

    DatasetBuilder class is responsible for loading, preprocessing, and transforming the dataset.
    The dataset is loaded from the path or generated by the defined method in the config.
    After loading the dataset, the raw preprocess method is applied to the dataset. This process is optional.
    Next step is splitting the dataset into train and test sets. The test set size is defined in the config.
    Finally, the transform method is applied to the dataset. This process is optional.
    builder ruturns the Dataset object that contains the train and test split of the dataset.

    Examples:
        >>> import numpy as np
        >>> from qxmt.configs import ExperimentConfig
        >>> from qxmt.datasets.builder import DatasetBuilder
        >>> config = ExperimentConfig(path="configs/my_run.yaml")
        >>> dataset = DatasetBuilder(config).build()
        Dataset(
            X_train=array([[15.81324596, -9.07999965], ...]),
            y_train=array([0, ...]),
            X_test=array([[6.71832813, 11.2653727], ...]),
            y_test=array([1, ...]),
            config=DatasetConfig(type='file', ...)
    """

    def __init__(self, config: ExperimentConfig, logger: Logger = LOGGER) -> None:
        """Initialize the DatasetBuilder.

        Args:
            config (ExperimentConfig): experiment config that loaded from the yaml file
            logger (Logger, optional): logger for output messages. Defaults to LOGGER.
        """
        self.config: ExperimentConfig = config
        self.task_type: str = config.global_settings.task_type
        self.random_seed: int = config.dataset.random_seed
        self.logger: Logger = logger
        self.custom_raw_preprocess_list = self._set_custom_raw_process_list()
        self.custom_transform_list = self._set_custom_transform_list()

    def _set_custom_raw_process_list(self) -> Optional[list[Callable]]:
        """Set the custom raw preprocess functions.
        If the custom raw preprocess functions are defined in the config, load and set them from top to bottom order.

        Returns:
            Optional[list[Callable]]: list of custom raw preprocess functions
        """
        raw_process_list = self.config.dataset.raw_preprocess_logic
        if (raw_process_list is None) or (len(raw_process_list) == 0):
            return None

        if isinstance(raw_process_list, dict):
            raw_process_list = [raw_process_list]

        custom_raw_process_list = []
        for logic in raw_process_list:
            raw_preprocess_logic = load_object_from_yaml(logic)
            self._validate_raw_preprocess_logic(raw_preprocess_logic, self.logger)
            custom_raw_process_list.append(raw_preprocess_logic)

        return custom_raw_process_list

    def _set_custom_transform_list(self) -> Optional[list[Callable]]:
        """Set the custom transform functions.
        If the custom transform functions are defined in the config, load and set them from top to bottom order.

        Returns:
            Optional[list[Callable]]: list of custom transform functions
        """
        transform_list = self.config.dataset.transform_logic
        if (transform_list is None) or (len(transform_list) == 0):
            return None

        if isinstance(transform_list, dict):
            transform_list = [transform_list]

        custom_transform_list = []
        for logic in transform_list:
            transform_logic = load_object_from_yaml(logic)
            self._validate_transform_logic(transform_logic, self.logger)
            custom_transform_list.append(transform_logic)

        return custom_transform_list

    @staticmethod
    def _validate_raw_preprocess_logic(raw_preprocess_logic: Callable, logger: Logger) -> None:
        """Validate the custom raw preprocess function.

        Args:
            raw_preprocess_logic (Callable): custom raw preprocess function
            logger (Logger): logger for output messages

        Raises:
            ValueError: argment lenght of the custom raw preprocess function is less than 2
            ValueError: return type of the custom raw preprocess function is not a tuple of numpy arrays
            ValueError: argument type of the custom raw preprocess function is not numpy array
        """
        type_hint_dict = get_type_hints(raw_preprocess_logic)
        parameter_dict = inspect.signature(raw_preprocess_logic).parameters

        # check argment length. -1 means return type
        if len(type_hint_dict) - 1 != len(parameter_dict):
            logger.warning(
                "All arguments of the custom raw preprocess function assigned to the type hint."
                "Input and return type validation will be skipped."
            )
            return
        elif len(type_hint_dict) - 1 < 2:
            raise ValueError("The custom raw preprocess function must have at least 2 arguments (X, y).")

        # check argument type and return type
        for arg_name, arg_type in type_hint_dict.items():
            if (arg_name == "return") and (arg_type != RAW_DATASET_TYPE):
                raise ValueError(
                    "The return type of the custom raw preprocess function must be a tuple of numpy arrays."
                )
            # [TODO]: Handle anthor data types
            # elif (arg_name != "return") and (arg_type != RAW_DATA_TYPE):
            #     raise ValueError(f'The arguments of the custom raw preprocess function must be "{RAW_DATA_TYPE}".')

    @staticmethod
    def _validate_transform_logic(transform_logic: Callable, logger: Logger) -> None:
        """Validate the custom transform function.

        Args:
            transform_logic (Callable): custom transform function
            logger (Logger): logger for output messages

        Raises:
            ValueError: argment lenght of the custom transform function is less than 6
            ValueError: return type of the custom transform function is not a tuple of numpy arrays
            ValueError: argument type of the custom transform function is not numpy array

        """
        type_hint_dict = get_type_hints(transform_logic)
        parameter_dict = inspect.signature(transform_logic).parameters

        # check argment length. -1 means return type
        if len(type_hint_dict) - 1 != len(parameter_dict):
            logger.warning(
                "All arguments of the custom raw preprocess function assigned to the type hint."
                "Input and return type validation will be skipped."
            )
            return
        elif len(type_hint_dict) - 1 < 6:
            raise ValueError(
                "The custom transform function must have at "
                "least 6 arguments (X_train, y_train, X_val, y_val, X_test, y_test)."
            )

        # check argument type and return type
        for arg_name, arg_type in type_hint_dict.items():
            if (arg_name == "return") and (arg_type != PROCESSCED_DATASET_TYPE):
                raise ValueError("The return type of the custom transform function must be a tuple of numpy arrays.")
            # # [TODO]: Handle athor data types
            # elif (arg_name != "return") and (arg_type != RAW_DATA_TYPE):
            #     raise ValueError(f'The arguments of the custom transform function must be "{RAW_DATA_TYPE}".')

    @staticmethod
    def _get_dataset_type(dataset_config: DatasetConfig) -> str:
        dataset_sources = []
        for source in ["openml", "file", "generate"]:
            if getattr(dataset_config, source) is not None:
                dataset_sources.append(source)

        if len(dataset_sources) != 1:
            raise ValueError("Exactly one of 'openml', 'file', or 'generate' must be set.")

        return dataset_sources[0]

    def load(self) -> RAW_DATASET_TYPE:
        """Load the dataset from the path defined in config.

        Returns:
            RAW_DATASET_TYPE: features and labels of the dataset
        """
        dataset_type = self._get_dataset_type(self.config.dataset)

        match dataset_type:
            case "openml":
                openml_config = cast(OpenMLConfig, self.config.dataset.openml)
                X, y = OpenMLDataLoader(
                    name=openml_config.name,
                    id=openml_config.id,
                    save_path=openml_config.save_path,
                    return_format=openml_config.return_format,
                ).load()
                X = cast(np.ndarray, X)
                y = cast(np.ndarray, y)
            case "file":
                file_config = cast(FileConfig, self.config.dataset.file)
                X, y = FileDataLoader(
                    data_path=file_config.data_path,
                    label_path=file_config.label_path,
                    label_name=file_config.label_name,
                ).load()
            case "generate":
                generate_config = cast(GenerateDataConfig, self.config.dataset.generate)
                X, y = GeneratedDataLoader(
                    task_type=self.task_type,
                    generate_method=generate_config.generate_method,
                    random_seed=self.random_seed,
                    params=generate_config.params or {},
                ).load()
            case _:
                raise ValueError(f"Invalid dataset type: {dataset_type}")

        return X, y

    def default_raw_preprocess(self, X: np.ndarray, y: np.ndarray) -> RAW_DATASET_TYPE:
        """Default raw preprocess method. This method does not apply any preprocess.

        Args:
            X (np.ndarray): raw features of the dataset
            y (np.ndarray): raw labels of the dataset

        Returns:
            RAW_DATASET_TYPE: raw features and labels of the dataset
        """
        return X, y

    def raw_preprocess(self, X: np.ndarray, y: np.ndarray) -> RAW_DATASET_TYPE:
        """Preprocess the raw dataset. This process executes before splitting the dataset.
        ex) filtering, data sampling, etc.

        Args:
            X (np.ndarray): raw features of the dataset
            y (np.ndarray): raw labels of the dataset

        Returns:
            RAW_DATASET_TYPE: preprocessed features and labels of the dataset
        """
        if self.custom_raw_preprocess_list is not None:
            for raw_preprocess_logic in self.custom_raw_preprocess_list:
                X, y = raw_preprocess_logic(X, y)
            return X, y
        else:
            return self.default_raw_preprocess(X, y)

    def split(self, X: np.ndarray, y: np.ndarray) -> PROCESSCED_DATASET_TYPE:
        """Split the dataset into train and test sets.
        Test set size is defined in the config.

        Args:
            X (np.ndarray): raw features of the dataset
            y (np.ndarray): raw labels of the dataset

        Returns:
            PROCESSCED_DATASET_TYPE: train and test split of dataset (features and labels)
        """
        split_config = self.config.dataset.split
        val_and_test_ratio = split_config.validation_ratio + split_config.test_ratio
        shuffle = split_config.shuffle

        # Split the dataset into train, validation, and test sets
        X_train, X_val_and_test, y_train, y_val_and_test = train_test_split(
            X,
            y,
            test_size=val_and_test_ratio,
            random_state=self.random_seed,
            shuffle=shuffle,
        )

        if split_config.validation_ratio == 0:
            # Validation set is not used
            X_val, y_val = None, None
            X_test, y_test = X_val_and_test, y_val_and_test
        else:
            # Split the validation and test sets
            X_val, X_test, y_val, y_test = train_test_split(
                X_val_and_test,
                y_val_and_test,
                test_size=split_config.test_ratio / val_and_test_ratio,
                random_state=self.random_seed,
                shuffle=shuffle,
            )

        return X_train, y_train, X_val, y_val, X_test, y_test

    def default_transform(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_val: Optional[np.ndarray],
        y_val: Optional[np.ndarray],
        X_test: np.ndarray,
        y_test: np.ndarray,
    ) -> PROCESSCED_DATASET_TYPE:
        """Default transform method. This method does not apply any transformation.

        Args:
            X_train (np.ndarray): raw features of the training data
            y_train (np.ndarray): raw labels of the training data
            X_val (Optional[np.ndarray]): raw features of the validation data. None if validation set is not used
            y_val (Optional[np.ndarray]): raw labels of the validation data. None if validation set is not used
            X_test (np.ndarray): raw features of the test data
            y_test (np.ndarray): raw labels of the test data

        Returns:
            PROCESSCED_DATASET_TYPE: train, val and test split of dataset (features and labels)
        """
        return X_train, y_train, X_val, y_val, X_test, y_test

    def transform(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_val: Optional[np.ndarray],
        y_val: Optional[np.ndarray],
        X_test: np.ndarray,
        y_test: np.ndarray,
    ) -> PROCESSCED_DATASET_TYPE:
        """Transform the dataset.
        ex) feature scaling, dimension reduction, etc.

        Args:
            X_train (np.ndarray): raw features of the training data
            y_train (np.ndarray): raw labels of the training data
            X_val (Optional[np.ndarray]): raw features of the validation data. None if validation set is not used
            y_val (Optional[np.ndarray]): raw labels of the validation data. None if validation set is not used
            X_test (np.ndarray): raw features of the test data
            y_test (np.ndarray): raw labels of the test data

        Returns:
            PROCESSCED_DATASET_TYPE: transformed train, val  and test split of dataset (features and labels)
        """
        if self.custom_transform_list is not None:
            for transform_logic in self.custom_transform_list:
                X_train, y_train, X_val, y_val, X_test, y_test = transform_logic(
                    X_train, y_train, X_val, y_val, X_test, y_test
                )
            return X_train, y_train, X_val, y_val, X_test, y_test
        else:
            return self.default_transform(X_train, y_train, X_val, y_val, X_test, y_test)

    def build(self) -> Dataset:
        """Build the dataset. This method loads, preprocesses, splits, and transforms the dataset.

        Returns:
            Dataset: Dataset object that contains the train and test split of the dataset
        """
        X, y = self.load()
        X, y = self.raw_preprocess(X, y)
        X_train, y_train, X_val, y_val, X_test, y_test = self.split(X, y)
        X_train_trs, y_train_trs, X_val_trs, y_val_trs, X_test_trs, y_test_trs = self.transform(
            X_train, y_train, X_val, y_val, X_test, y_test
        )

        return Dataset(
            X_train=X_train_trs,
            y_train=y_train_trs,
            X_val=X_val_trs,
            y_val=y_val_trs,
            X_test=X_test_trs,
            y_test=y_test_trs,
            config=self.config.dataset,
        )



from typing import Any, Optional

import numpy as np

from qxmt.exceptions import InvalidPlatformError


class BaseDevice:
    """General-purpose device class for experiment.
    This class is abstracted to oversee multiple platforms.
    Provide a common interface within the QXMT library by absorbing differences between platforms.

    Examples:
        >>> from qxmt.devices.base import BaseDevice
        >>> device = BaseDevice(platform="pennylane", name="default.qubit", n_qubits=2, shots=100)
        >>> simulator = device.get_simulator()
    """

    def __init__(
        self, platform: str, name: str, n_qubits: int, shots: Optional[int], random_seed: Optional[int] = None
    ) -> None:
        """Initialize the quantum device.

        Args:
            platform (str): platform name (ex: pennylane, qulacs, etc.)
            name (str): device name provided by the platform (ex: default.qubit, default.tensor, etc.)
            n_qubits (int): number of qubits
            shots (Optional[int]): number of shots for the quantum circuit
            random_seed (Optional[int]): random seed for the quantum device
        """
        self.platform = platform
        self.name = name
        self.n_qubits = n_qubits
        self.shots = shots
        self.random_seed = random_seed

    def get_simulator(self) -> Any:
        """Get quantum simulator.

        Raises:
            InvalidPlatformError: platform is not implemented.
        """
        if self.platform == "pennylane":
            import pennylane as qml

            return qml.device(
                name=self.name,
                wires=self.n_qubits,
                shots=self.shots,
                seed=np.random.default_rng(self.random_seed) if self.random_seed is not None else None,
            )
        else:
            raise InvalidPlatformError(f'"{self.platform}" is not implemented.')

    def get_real_machine(self) -> Any:
        """Get quantum real machine.

        Raises:
            InvalidPlatformError: platform is not implemented.
        """
        raise NotImplementedError



from qxmt.evaluation.evaluation import (
    ClassificationEvaluation,
    Evaluation,
    RegressionEvaluation,
)

__all__ = [
    "Evaluation",
    "ClassificationEvaluation",
    "RegressionEvaluation",
]



from qxmt.configs import (
    DatasetConfig,
    DeviceConfig,
    EvaluationConfig,
    ExperimentConfig,
    FeatureMapConfig,
    FileConfig,
    GenerateDataConfig,
    GlobalSettingsConfig,
    KernelConfig,
    ModelConfig,
    SplitConfig,
)
from qxmt.exceptions import (
    DeviceSettingError,
    ExperimentNotInitializedError,
    ExperimentRunSettingError,
    ExperimentSettingError,
    InputShapeError,
    InvalidFileExtensionError,
    InvalidModelNameError,
    InvalidPlatformError,
    InvalidQunatumDeviceError,
    JsonEncodingError,
    ModelSettingError,
    ReproductionError,
)
from qxmt.experiment.experiment import Experiment

__all__ = [
    "DatasetConfig",
    "DeviceConfig",
    "EvaluationConfig",
    "ExperimentConfig",
    "FeatureMapConfig",
    "FileConfig",
    "GenerateDataConfig",
    "GlobalSettingsConfig",
    "KernelConfig",
    "ModelConfig",
    "SplitConfig",
    "DeviceSettingError",
    "ExperimentNotInitializedError",
    "ExperimentRunSettingError",
    "ExperimentSettingError",
    "InputShapeError",
    "InvalidFileExtensionError",
    "InvalidModelNameError",
    "InvalidPlatformError",
    "InvalidQunatumDeviceError",
    "JsonEncodingError",
    "ModelSettingError",
    "ReproductionError",
    "Experiment",
]


__version__ = "0.3.7"



import numpy as np
import pennylane as qml

from qxmt.constants import PENNYLANE_PLATFORM
from qxmt.feature_maps.base import BaseFeatureMap


class XXFeatureMap(BaseFeatureMap):
    """XX feature map class.

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.ising import XXFeatureMap
        >>> feature_map = XXFeatureMap(2, 2)
        >>> feature_map.draw(x_dim=2)
        0: ──RX(0.33)─╭IsingXX(15.19)──RX(0.33)─╭IsingXX(15.19)─┤
        1: ──RX(0.44)─╰IsingXX(15.19)──RX(0.44)─╰IsingXX(15.19)─┤
    """

    def __init__(self, n_qubits: int, reps: int) -> None:
        """Initialize the XX feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.reps: int = reps

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of XX feature map.

        Args:
            x (np.ndarray): input data
        """
        for _ in range(self.reps):
            for i in range(self.n_qubits):
                qml.RX(x[i], wires=i)
            for i in range(0, self.n_qubits - 1):
                qml.IsingXX(2 * (np.pi - x[i]) * (np.pi - x[i + 1]), wires=[i, i + 1])


class YYFeatureMap(BaseFeatureMap):
    """YY feature map class.

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.ising import YYFeatureMap
        >>> feature_map = YYFeatureMap(2, 2)
        >>> feature_map.draw(x_dim=2)
        0: ──H──RY(0.86)─╭IsingYY(11.23)──H──RY(0.86)─╭IsingYY(11.23)─┤
        1: ──H──RY(0.68)─╰IsingYY(11.23)──H──RY(0.68)─╰IsingYY(11.23)─┤
    """

    def __init__(self, n_qubits: int, reps: int) -> None:
        """Initialize the YY feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.reps: int = reps

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of YY feature map.

        Args:
            x (np.ndarray): input data
        """
        for _ in range(self.reps):
            for i in range(self.n_qubits):
                qml.Hadamard(wires=i)
                qml.RY(x[i], wires=i)
            for i in range(0, self.n_qubits - 1):
                qml.IsingYY(2 * (np.pi - x[i]) * (np.pi - x[i + 1]), wires=[i, i + 1])


class ZZFeatureMap(BaseFeatureMap):
    """ZZ feature map class.

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.ising import ZZFeatureMap
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> feature_map.draw(x_dim=2)
        0: ──H──RZ(0.88)─╭IsingZZ(13.28)──H──RZ(0.88)─╭IsingZZ(13.28)─┤
        1: ──H──RZ(0.20)─╰IsingZZ(13.28)──H──RZ(0.20)─╰IsingZZ(13.28)─┤
    """

    def __init__(self, n_qubits: int, reps: int) -> None:
        """Initialize the ZZ feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.reps: int = reps

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of ZZ feature map.

        Args:
            x (np.ndarray): input data
        """
        for _ in range(self.reps):
            for i in range(self.n_qubits):
                qml.Hadamard(wires=i)
                qml.RZ(x[i], wires=i)
            for i in range(0, self.n_qubits - 1):
                qml.IsingZZ(2 * (np.pi - x[i]) * (np.pi - x[i + 1]), wires=[i, i + 1])



import copy
from pathlib import Path
from typing import Any, Callable, Optional

import dill
import numpy as np
from sklearn.model_selection import cross_val_score
from sklearn.svm import SVR

from qxmt.constants import DEFAULT_N_JOBS
from qxmt.kernels.base import BaseKernel
from qxmt.models.base import BaseKernelModel
from qxmt.models.hyperparameter_search.search import HyperParameterSearch


class QSVR(BaseKernelModel):
    """Quantum Support Vector Regression (QSVR) model.
    This class wraps the sklearn.svm.SVR class to provide a QSVR model.
    Then, many methods use the same interface as the sklearn.svm.SVR class.
    References: https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html

    Examples:
        >>> from qxmt.models.qsvc import QSVR
        >>> from qxmt.kernels.pennylane import FidelityKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1000,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> kernel = FidelityKernel(device, feature_map)
        >>> model = QSVR(kernel=kernel)
        >>> model.fit(X_train, y_train)
        >>> model.predict(X_test)
        np.array([-0.5441202 , -0.19483975, -0.19773099, -0.24928479,  0.12222858])
    """

    def __init__(self, kernel: BaseKernel, n_jobs: int = DEFAULT_N_JOBS, **kwargs: Any) -> None:
        """Initialize the QSVR model.

        Args:
            kernel (BaseKernel): kernel instance of BaseKernel class
            n_jobs (int): number of jobs for parallel computation
        """
        super().__init__(kernel, n_jobs)
        self.fit_X: Optional[np.ndarray] = None
        self.model = SVR(kernel="precomputed", **kwargs)

    def __getattr__(self, name: str) -> Any:
        # if the attribute is in the model, return it
        if hasattr(self.model, name):
            return getattr(self.model, name)
        # if the attribute is not in the model, raise AttributeError
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def cross_val_score(
        self,
        X: np.ndarray,
        y: np.ndarray,
        **kwargs: Any,
    ) -> np.ndarray:
        """Cross validation score of the QSVR model.
        Default to use the R^2 score.

        Args:
            X (np.ndarray): numpy array of features
            y (np.ndarray): numpy array of target values

        Returns:
            np.ndarray: numpy array of scores
        """
        kernel_X, _ = self.kernel.compute_matrix(X, X, return_shots_resutls=False, n_jobs=self.n_jobs)
        scores = cross_val_score(estimator=self.model, X=kernel_X, y=y, n_jobs=self.n_jobs, **kwargs)

        return scores

    def hyperparameter_search(
        self,
        X: np.ndarray,
        y: np.ndarray,
        search_type: str,
        search_space: dict[str, list[Any]],
        search_args: dict[str, Any],
        objective: Optional[Callable] = None,
        refit: bool = True,
    ) -> dict[str, Any]:
        """Search the best hyperparameters for the QSVR model.

        Args:
            X (np.ndarray): dataset for search
            y (np.ndarray): target values for search
            search_type (str): search type for hyperparameter search (grid, random, tpe)
            search_space (dict[str, list[Any]]): search space for hyperparameter search
            search_args (dict[str, Any]): search arguments for hyperparameter search
            objective (Optional[Callable], optional): objective function for search. Defaults to None.
            refit (bool, optional): refit the model with best hyperparameters. Defaults to True.

        Raises:
            ValueError: Not supported search type

        Returns:
            dict[str, Any]: best hyperparameters
        """
        search_model = copy.deepcopy(self.model)
        X_kernel, _ = self.kernel.compute_matrix(X, X, return_shots_resutls=False, n_jobs=self.n_jobs)

        if "scoring" not in search_args.keys():
            search_args["scoring"] = "r2"

        searcher = HyperParameterSearch(
            X=X_kernel,
            y=y,
            model=search_model,
            search_type=search_type,
            search_space=search_space,
            search_args=search_args,
            objective=objective,
        )
        best_params = searcher.search()

        if refit:
            self.model.set_params(**best_params)
            self.model.fit(X_kernel, y)

        return best_params

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        save_shots_path: Optional[Path | str] = None,
        **kwargs: Any,
    ) -> None:
        """Fit the model with given features and target values.

        Args:
            X (np.ndarray): numpy array of features
            y (np.ndarray): numpy array of target values
            save_shots_path (Optional[Path | str], optional): save path for the shot results. Defaults to None.
        """
        self.fit_X = X
        if save_shots_path is not None:
            kernel_train_X, shots_matrix = self.kernel.compute_matrix(
                self.fit_X, self.fit_X, return_shots_resutls=True, n_jobs=self.n_jobs, bar_label="Train"
            )
            if shots_matrix is not None:
                self.kernel.save_shots_results(shots_matrix, save_shots_path)
        else:
            kernel_train_X, _ = self.kernel.compute_matrix(
                self.fit_X, self.fit_X, return_shots_resutls=False, n_jobs=self.n_jobs, bar_label="Train"
            )
        self.model.fit(kernel_train_X, y, **kwargs)

    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict the target value with given features.

        Args:
            X (np.ndarray): numpy array of features

        Returns:
            np.ndarray: numpy array of predicted values
        """
        if self.fit_X is None:
            raise ValueError("The model is not trained yet.")
        else:
            kernel_pred_X, _ = self.kernel.compute_matrix(
                X, self.fit_X, return_shots_resutls=False, n_jobs=self.n_jobs, bar_label="Test"
            )
        return self.model.predict(kernel_pred_X)

    def save(self, path: str | Path) -> None:
        """Save the model to the given path.

        Args:
            path (str | Path): path to save the model
        """
        # [TODO] Use pickle of joblib
        # AttributeError: Can't pickle local object 'BaseKernel._to_fm_instance.<locals>.CustomFeatureMap'
        dill.dump(self.model, open(path, "wb"))

    def load(self, path: str | Path) -> "QSVR":
        """Load the trained model from the given path.

        Args:
            path (str | Path): path to load the model

        Returns:
            QSVR: loaded QSVC model
        """
        # [TODO] Use pickle of joblib
        return dill.load(open(path, "rb"))

    def get_params(self) -> dict:
        """Get the parameters of the model."""
        return self.model.get_params()

    def set_params(self, params: dict) -> None:
        """Set the parameters of the model."""
        self.model.set_params(**params)



from qxmt.evaluation.metrics.base import BaseMetric
from qxmt.evaluation.metrics.defaults_classification import (
    Accuracy,
    F1Score,
    Precision,
    Recall,
)
from qxmt.evaluation.metrics.defaults_regression import (
    MeanAbsoluteError,
    R2Score,
    RootMeanSquaredError,
)

__all__ = [
    "BaseMetric",
    "Accuracy",
    "Recall",
    "Precision",
    "F1Score",
    "MeanAbsoluteError",
    "RootMeanSquaredError",
    "R2Score",
]



from typing import Callable, Literal, cast

import numpy as np
import pennylane as qml
from pennylane.measurements.probs import ProbabilityMP
from pennylane.measurements.sample import SampleMP

from qxmt.devices.base import BaseDevice
from qxmt.exceptions import ModelSettingError
from qxmt.feature_maps.base import BaseFeatureMap
from qxmt.kernels.base import BaseKernel
from qxmt.kernels.sampling import sample_results_to_probs


class ProjectedKernel(BaseKernel):
    """Projected kernel class.
    The projected kernel is a quantum kernel that projects the quantum state to a specific basis
    and computes the kernel value based on the projected measurement results.
    Reference: https://www.nature.com/articles/s41467-021-22539-9

    Args:
        BaseKernel (_type_): base class of kernel

    Examples:
        >>> import numpy as np
        >>> from qxmt.kernels.pennylane.projected_kernel import ProjectedKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1024,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> kernel = ProjectedKernel(device, feature_map)
        >>> x1 = np.random.rand(2)
        >>> x2 = np.random.rand(2)
        >>> kernel.compute(x1, x2)
        0.86
    """

    def __init__(
        self,
        device: BaseDevice,
        feature_map: BaseFeatureMap | Callable[[np.ndarray], None],
        gamma: float = 1.0,
        projection: Literal["x", "y", "z"] = "z",
    ) -> None:
        """Initialize the ProjectedKernel class.

        Args:
            device (BaseDevice): device instance for quantum computation
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function
            gamma (float): gamma parameter for kernel computation
            projection (str): projection method for kernel computation
        """
        if projection not in ["x", "y", "z"]:
            raise ValueError("Projection method must be 'x', 'y', or 'z'.")

        super().__init__(device, feature_map)
        self.n_qubits = device.n_qubits
        self.gamma = gamma
        self.projection = projection

    def _process_measurement_results(self, results: list[float]) -> np.ndarray:
        """Process the measurement results based on the projection method.

        Args:
            results (list[float]): projected measurement results for each qubit

        Returns:
            np.ndarray: processed measurement results
        """
        if self.projection == "xyz_sum":
            sum_results: list[float] = []
            idx_size = 3  # x, y, z
            for i in range(0, self.n_qubits, idx_size):
                sum_result: float = np.sum([result for result in results[i : i + idx_size]])
                sum_results.append(sum_result)
            projected_results = np.array(sum_results)
        else:
            projected_results = np.array(results)

        return projected_results

    def _calculate_expected_values_by_z(self, probs: np.ndarray, target_qubit: int) -> float:
        mask = 1 << target_qubit
        expval_z = 0.0
        for i, prob in enumerate(probs):
            if (i & mask) == 0:
                expval_z += prob
            else:
                expval_z -= prob
        return expval_z

    def _calculate_expected_values(self, probs: np.ndarray) -> np.ndarray:
        # calculate the expected values by Z basis for each qubit
        # when the projection method is "x" or "y", apply Hadamard or RY gate in _circuit method
        projected_exp_value = np.array([self._calculate_expected_values_by_z(probs, i) for i in range(self.n_qubits)])

        return projected_exp_value

    def _circuit(self, x: np.ndarray) -> ProbabilityMP | SampleMP:
        if self.feature_map is None:
            raise ModelSettingError("Feature map must be provided for FidelityKernel.")

        self.feature_map(x)

        # apply projection operators for calculating the expected values by Z basis
        if self.projection == "x":
            for i in range(self.n_qubits):
                qml.Hadamard(wires=i)
        elif self.projection == "y":
            for i in range(self.n_qubits):
                qml.RY(np.pi / 2, wires=i)

        if self.is_sampling:
            return qml.sample(wires=range(self.n_qubits))
        else:
            return qml.probs(wires=range(self.n_qubits))

    def compute(self, x1: np.ndarray, x2: np.ndarray) -> tuple[float, np.ndarray]:
        """Compute the projected kernel value between two data points.

        Args:
            x1 (np.ndarray): numpy array representing the first data point
            x2 (np.ndarray): numpy array representing the second data point

        Returns:
            tuple[float, np.ndarray]: projected kernel value and probability distribution
        """
        qnode = qml.QNode(self._circuit, device=self.device.get_simulator(), cache=False)  # type: ignore
        x1_result = qnode(x1)
        x2_result = qnode(x2)

        if self.is_sampling:
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            x1_probs = sample_results_to_probs(x1_result, self.n_qubits, cast(int, self.device.shots))
            x2_probs = sample_results_to_probs(x2_result, self.n_qubits, cast(int, self.device.shots))
        else:
            # use theoretical probability distribution
            x1_probs = np.array(x1_result)
            x2_probs = np.array(x2_result)

        # compute expected values for projection operators
        x1_projected = self._calculate_expected_values(x1_probs)
        x2_projected = self._calculate_expected_values(x2_probs)

        # compute gaussian kernel value based on the projected measurement results
        kernel_value = np.exp(-self.gamma * np.sum((x1_projected - x2_projected) ** 2))

        return kernel_value, (x1_probs + x2_probs) / 2



from pathlib import Path

from pydantic import BaseModel, ConfigDict

from qxmt.datasets.schema import Dataset
from qxmt.models.base import BaseMLModel


class RunTime(BaseModel):
    fit_seconds: float
    predict_seconds: float


class RunRecord(BaseModel):
    model_config = ConfigDict(frozen=True, extra="forbid")

    run_id: int
    desc: str
    execution_time: str
    runtime: RunTime
    commit_id: str
    config_file_name: Path
    evaluation: dict[str, float]


class RunArtifact(BaseModel):
    model_config = ConfigDict(frozen=True, arbitrary_types_allowed=True, extra="forbid")

    run_id: int
    dataset: Dataset
    model: BaseMLModel


class ExperimentDB(BaseModel):
    model_config = ConfigDict(extra="forbid")

    name: str
    desc: str
    working_dirc: Path
    experiment_dirc: Path
    runs: list[RunRecord]



from typing import Any, Optional, Type, get_args

import numpy as np
import pandas as pd

from qxmt.evaluation.metrics.base import BaseMetric
from qxmt.evaluation.metrics.defaults_classification import (
    DEFAULT_CLF_METRICS_NAME,
    NAME2CLF_METRIC,
)
from qxmt.evaluation.metrics.defaults_regression import (
    DEFAULT_REG_METRICS_NAME,
    NAME2REG_METRIC,
)
from qxmt.utils import load_object_from_yaml


class Evaluation:
    """Evaluation class for model evaluation.
    This class is used to evaluate the model performance by comparing the actual and predicted values.
    Evaluation metrics are defined by default and custom metrics.
    Furthermore, the result can be accessed as a dictionary or DataFrame.

    Examples:
        >>> from qxmt.evaluation.evaluation import Evaluation
        >>> actual = np.array([1, 0, 1])
        >>> predicted = np.array([1, 1, 1])
        >>> evaluation = Evaluation(actual, predicted)
        >>> evaluation.evaluate()
        >>> evaluation.to_dict()
        {'accuracy': 0.6666666666666666, 'precision': 0.6666666666666666, 'recall': 1.0, 'f1_score': 0.8}
        >>> evaluation.to_dataframe()
           accuracy  precision  recall  f1_score
        0  0.666667   0.666667     1.0       0.8
    """

    DEFAULT_METRICS_NAME: list[str] = []
    NAME2METRIC: dict[str, Type[BaseMetric]] = {}

    def __init__(
        self,
        actual: np.ndarray,
        predicted: np.ndarray,
        default_metrics_name: Optional[list[str]] = None,
        custom_metrics: Optional[list[dict[str, Any]]] = None,
    ) -> None:
        """Initialize the evaluation class.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values
            default_metrics_name (Optional[list[str]], optional):
                metrics name list defined by default. Defaults to None.
            custom_metrics (Optional[list[dict[str, Any]]], optional):
                metrics name list defined by user custom. Defaults to None.
        """
        self.actual: np.ndarray = actual
        self.predicted: np.ndarray = predicted
        self.default_metrics_name: list[str]
        self.custom_metrics_name: list[str]
        self.default_metrics: list[BaseMetric]
        self.custom_metrics: list[BaseMetric]

        self.init_default_metrics(default_metrics_name)
        self.init_custom_metrics(custom_metrics)

    def __repr__(self) -> str:
        return (
            f"Evaluation(actual={self.actual.tolist()}, "
            f"predicted={self.predicted.tolist()}, "
            f"default_metrics_name={self.default_metrics_name})"
            f"custom_metrics_name={self.custom_metrics_name}"
        )

    def init_default_metrics(self, default_metrics_name: Optional[list[str]]) -> None:
        """Initialize and validate default metrics.

        Raises:
            ValueError: if the metric is not implemented
        """
        if default_metrics_name is not None:
            self.default_metrics_name = default_metrics_name
        else:
            self.default_metrics_name = self.DEFAULT_METRICS_NAME

        self.default_metrics = []
        for metric_name in self.default_metrics_name:
            if metric_name not in self.NAME2METRIC:
                raise ValueError(f"{metric_name} is not implemented.")

            metric = self.NAME2METRIC[metric_name](name=metric_name)
            self.default_metrics.append(metric)

    def init_custom_metrics(self, custom_metrics: Optional[list[dict[str, Any]]]) -> None:
        """Initialize and validate custom metrics.

        Args:
            custom_metrics (Optional[list[dict[str, Any]]]): list of custom metrics. Defaults to None.

        Raises:
            ValueError: if the metric is not subclass of BaseMetric
        """
        self.custom_metrics_name = []
        self.custom_metrics = []

        if custom_metrics is not None:
            for metric_config in custom_metrics:
                metric = load_object_from_yaml(config=metric_config)
                if not isinstance(metric, BaseMetric):
                    raise ValueError("Custom metrics must be a subclass of BaseMetric.")

                self.custom_metrics.append(metric)
                self.custom_metrics_name.append(metric.name)

    def set_evaluation_result(self, metrics: list[BaseMetric]) -> None:
        """Evaluate default metrics.

        Raises:
            ValueError: if the metric is not implemented
        """
        for metric in metrics:
            metric.set_score(self.actual, self.predicted)

    def evaluate(self) -> None:
        """Evaluate default and custom metrics."""
        self.set_evaluation_result(self.default_metrics)
        self.set_evaluation_result(self.custom_metrics)

    def to_dict(self) -> dict:
        """Convert evaluation metrics to dictionary.

        Raises:
            ValueError: if the metrics are not evaluated yet

        Returns:
            dict: dictionary of evaluation metrics
        """
        metrics = self.default_metrics + self.custom_metrics
        for metric in metrics:
            if metric.score is None:
                raise ValueError("Metrics are not evaluated yet.")

        data = {metric.name: metric.score for metric in metrics}

        return data

    def to_dataframe(
        self,
        id: Optional[str] = None,
        id_columns_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """Convert evaluation metrics to DataFrame.

        Args:
            id (Optional[str], optional): id of the evaluation (ex: run_id). Defaults to None.
            id_columns_name (Optional[str], optional): name of the id column. Defaults to None.

        Returns:
            pd.DataFrame: DataFrame of evaluation metrics
        """
        data = self.to_dict()
        df = pd.DataFrame(data, index=[0])
        if (id is not None) and (id_columns_name is not None):
            df.insert(0, id_columns_name, id)

        return df


class ClassificationEvaluation(Evaluation):
    DEFAULT_METRICS_NAME = list(get_args(DEFAULT_CLF_METRICS_NAME))
    NAME2METRIC = NAME2CLF_METRIC


class RegressionEvaluation(Evaluation):
    DEFAULT_METRICS_NAME = list(get_args(DEFAULT_REG_METRICS_NAME))
    NAME2METRIC = NAME2REG_METRIC



import os

if os.getenv("USE_LLM", "FALSE").lower() == "true":
    import torch  # type: ignore
    from transformers import AutoModelForCausalLM, AutoTokenizer  # type: ignore

from qxmt.constants import LLM_MODEL_PATH
from qxmt.generators.prompts import diff_desc_system_prompt, diff_desc_user_prompt


class DescriptionGenerator:
    """Description generator class.
    This class is used to generate the description of each run.
    The description is generated by a large language model (LLM) based on the diff code of each commit.
    If use this class, set the environment variable "USE_LLM" to "TRUE".
    Otherwise, the class is not used even if it is imported or model exists.

    Examples:
        >>> from qxmt.generators.description import DescriptionGenerator
        >>> description_generator = DescriptionGenerator()
        >>> description_generator.generate(add_code="import numpy as np", remove_code="import pandas as pd")
        'The code imports numpy as np and removes import pandas as pd.'
    """

    def __init__(self, model_path: str = LLM_MODEL_PATH) -> None:
        """Initialize the description generator

        Args:
            model_path (str, optional): model name of llm for generate description. Defaults to LLM_MODEL_PATH.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.model = AutoModelForCausalLM.from_pretrained(model_path)

    def _create_message(self, system_prompt: str, user_prompt: str) -> list[dict]:
        """Create a message for the model input.

        Args:
            system_prompt (str): system prompt
            user_prompt (str): user prompt

        Returns:
            list[dict]: message for the model input
        """
        message = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": user_prompt},
        ]
        return message

    def generate(
        self,
        add_code: str = "",
        remove_code: str = "",
        system_prompt: str = diff_desc_system_prompt,
        user_prompt: str = diff_desc_user_prompt,
    ) -> str:
        """Generate a description for the input diff code.

        Args:
            add_code (str, optional): added code. Defaults to "".
            remove_code (str, optional): removed code. Defaults to "".
            system_prompt (str, optional): system prompt. Defaults to diff_desc_system_prompt.
            user_prompt (str, optional): user prompt. Defaults to diff_desc_user_prompt.

        Returns:
            str: generated description
        """
        if add_code == "" and remove_code == "":
            return "No code changes detected on local git repository."

        user_prompt = user_prompt.format(add_code=add_code, remove_code=remove_code)
        message = self._create_message(system_prompt, user_prompt)
        prompt = self.tokenizer.apply_chat_template(message, tokenize=False, add_generation_prompt=True)
        token_ids: torch.Tensor = self.tokenizer.encode(
            str(prompt),
            add_special_tokens=False,
            return_tensors="pt",
        )  # type: ignore

        with torch.no_grad():
            output_ids = self.model.generate(
                token_ids.to(device=self.model.device),
                do_sample=None,
                temperature=0.0,
                max_new_tokens=256,
            )
        output = self.tokenizer.decode(output_ids.tolist()[0][token_ids.size(1) :], skip_special_tokens=True)

        return output



from typing import Optional, TypeVar

import numpy as np
import pennylane as qml

# [TODO]: constract from qxmt.constants
QuantumDeviceType = TypeVar("QuantumDeviceType", qml.devices.Device, qml.devices.LegacyDevice, qml.devices.QubitDevice)


RAW_DATA_TYPE = np.ndarray
RAW_LABEL_TYPE = np.ndarray
RAW_DATASET_TYPE = tuple[RAW_DATA_TYPE, RAW_LABEL_TYPE]
PROCESSCED_DATASET_TYPE = tuple[
    np.ndarray, np.ndarray, Optional[np.ndarray], Optional[np.ndarray], np.ndarray, np.ndarray
]



from qxmt.devices.base import BaseDevice
from qxmt.devices.builder import DeviceBuilder
from qxmt.devices.device_info import get_number_of_qubits, get_platform_from_device

__all__ = ["DeviceBuilder", "BaseDevice", "get_number_of_qubits", "get_platform_from_device"]



from typing import Any, Literal, Type

import numpy as np
from sklearn.metrics import mean_absolute_error, r2_score, root_mean_squared_error

from qxmt.evaluation.metrics.base import BaseMetric


class MeanAbsoluteError(BaseMetric):
    """Mean absolute error metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import MeanAbsoluteError
        >>> metric = MeanAbsoluteError()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        mean_absolute_error: 0.33
    """

    def __init__(self, name: str = "mean_absolute_error") -> None:
        """Initialize the mean absolute error metric.

        Args:
            name (str, optional): name of mean absolute error metric. Defaults to "mean_absolute_error".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the mean absolute error.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values

        Returns:
            float: mean absolute error score
        """
        score = mean_absolute_error(y_true=actual, y_pred=predicted, **kwargs)
        return float(score)


class RootMeanSquaredError(BaseMetric):
    """Root mean squared error metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import RootMeanSquaredError
        >>> metric = RootMeanSquaredError()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        root_mean_squared_error: 0.58
    """

    def __init__(self, name: str = "root_mean_squared_error") -> None:
        """Initialize the root mean squared error metric.

        Args:
            name (str, optional): name of root mean squared error metric. Defaults to "root_mean_squared_error".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the root mean squared error.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values

        Returns:
            float: root mean squared error score
        """
        score = root_mean_squared_error(y_true=actual, y_pred=predicted, **kwargs)
        return float(score)


class R2Score(BaseMetric):
    """R2 score metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import R2Score
        >>> metric = R2Score()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        r2_score: -0.5
    """

    def __init__(self, name: str = "r2_score") -> None:
        """Initialize the R2 score metric.

        Args:
            name (str, optional): name of R2 score metric. Defaults to "r2_score".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the R2 score.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values

        Returns:
            float: R2 score
        """
        score = r2_score(y_true=actual, y_pred=predicted, **kwargs)
        return float(score)


# set default regression metrics name list for evaluation
DEFAULT_REG_METRICS_NAME = Literal[
    "mean_absolute_error",
    "root_mean_squared_error",
    "r2_score",
]

NAME2REG_METRIC: dict[str, Type[BaseMetric]] = {
    "mean_absolute_error": MeanAbsoluteError,
    "root_mean_squared_error": RootMeanSquaredError,
    "r2_score": R2Score,
}



from typing import Any, Literal, Type

import numpy as np
from sklearn.metrics import accuracy_score, f1_score, precision_score, recall_score

from qxmt.evaluation.metrics.base import BaseMetric


class Accuracy(BaseMetric):
    """Accuracy metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import Accuracy
        >>> metric = Accuracy()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        accuracy: 0.67
    """

    def __init__(self, name: str = "accuracy") -> None:
        """Initialize the accuracy metric.

        Args:
            name (str, optional): name of accuracy metric. Defaults to "accuracy".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the accuracy score.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values

        Returns:
            float: accuracy score
        """
        score = accuracy_score(y_true=actual, y_pred=predicted, **kwargs)

        return float(score)


class Recall(BaseMetric):
    """Recall metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import Recall
        >>> metric = Recall()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        recall: 1.0
    """

    def __init__(self, name: str = "recall") -> None:
        """Initialize the recall metric.

        Args:
            name (str, optional): name of recall metric. Defaults to "recall".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the recall score.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values
            **kwargs (dict): additional keyword arguments. The following options are supported:
                - average (str): define averaging method
                - pos_label (int): positive label for binary classification (default: max value)

        Returns:
            float: recall score
        """
        if kwargs.get("average") is None:
            average = "binary" if len(np.unique(actual)) == 2 else "macro"
            kwargs["average"] = average

        # skleran default pos_label is 1.
        # it is handle not enclude 1 label case. ex) [0, 8] -> pos_label=8
        if (kwargs.get("average") == "binary") and (kwargs.get("pos_label") is None):
            pos_label = max(np.unique(actual))
            kwargs["pos_label"] = pos_label

        score = recall_score(y_true=actual, y_pred=predicted, **kwargs)

        return float(score)


class Precision(BaseMetric):
    """Precision metric class.

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import Precision
        >>> metric = Precision()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        precision: 0.67
    """

    def __init__(self, name: str = "precision") -> None:
        """Initialize the precision metric.

        Args:
            name (str, optional): name of precision metric. Defaults to "precision".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the precision score.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values
            **kwargs (dict): additional keyword arguments. The following options are supported:
                - average (str): define averaging method
                - pos_label (int): positive label for binary classification (default: max value)

        Returns:
            float: precision score
        """
        if kwargs.get("average") is None:
            average = "binary" if len(np.unique(actual)) == 2 else "macro"
            kwargs["average"] = average

        # skleran default pos_label is 1.
        # it is handle not enclude 1 label case. ex) [0, 8] -> pos_label=8
        if (kwargs.get("average") == "binary") and (kwargs.get("pos_label") is None):
            pos_label = max(np.unique(actual))
            kwargs["pos_label"] = pos_label

        score = precision_score(y_true=actual, y_pred=predicted, **kwargs)

        return float(score)


class F1Score(BaseMetric):
    """F1 score metric

    Args:
        BaseMetric (_type_): Base metric class

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.defaults import F1Score
        >>> metric = F1Score()
        >>> metric.set_score(np.array([1, 0, 1]), np.array([1, 1, 1]))
        >>> metric.output_score()
        f1_score: 0.8
    """

    def __init__(self, name: str = "f1_score") -> None:
        """Initialize the F1 score metric.

        Args:
            name (str, optional): name of f1-score metric. Defaults to "f1_score".
        """
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """Evaluate the F1 score.

        Args:
            actual (np.ndarray): numpy array of actual values
            predicted (np.ndarray): numpy array of predicted values
            **kwargs (dict): additional keyword arguments. The following options are supported:
                - average (str): define averaging method
                - pos_label (int): positive label for binary classification (default: max value)

        Returns:
            float: F1 score
        """
        if kwargs.get("average") is None:
            average = "binary" if len(np.unique(actual)) == 2 else "macro"
            kwargs["average"] = average

        # skleran default pos_label is 1.
        # it is handle not enclude 1 label case. ex) [0, 8] -> pos_label=8
        if (kwargs.get("average") == "binary") and (kwargs.get("pos_label") is None):
            pos_label = max(np.unique(actual))
            kwargs["pos_label"] = pos_label

        score = f1_score(y_true=actual, y_pred=predicted, **kwargs)

        return float(score)


class TargetAlignmet(BaseMetric):
    def __init__(self, name: str = "target_alignment") -> None:
        super().__init__(name)

    @staticmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        raise NotImplementedError


# set default classification metrics name list for evaluation
DEFAULT_CLF_METRICS_NAME = Literal[
    "accuracy",
    "precision",
    "recall",
    "f1_score",
]

NAME2CLF_METRIC: dict[str, Type[BaseMetric]] = {
    "accuracy": Accuracy,
    "precision": Precision,
    "recall": Recall,
    "f1_score": F1Score,
}



from pathlib import Path
from typing import Any, Optional

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from qxmt.datasets.schema import Dataset

DEFAULT_FEATURE_COLS = ["feature_1", "feature_2"]


def plot_2d_dataset(
    dataset: Dataset,
    colors: Optional[dict[int, tuple]] = None,
    class_labels: Optional[dict[int, str]] = None,
    save_path: Optional[str | Path] = None,
    **kwargs: Any,
) -> None:
    """Ploat dataset on 2D plane.

    Args:
        dataset (Dataset): Dataset object. It contains train and test data.
        colors (Optional[dict[int, tuple]], optional): color of each class. Defaults to None.
        class_labels (Optional[dict[int, str]], optional): label of each class. Defaults to None.
        save_path (Optional[str], optional): save path of graph. Defaults to None.
        **kwargs (Any): additional arguments for plot.
            train_title (str, optional): title of train dataset. Defaults to "Train Dataset".
            test_title (str, optional): title of test dataset. Defaults to "Test Dataset".
    """
    if dataset.config.features is not None:
        feature_cols = dataset.config.features
    else:
        feature_cols = DEFAULT_FEATURE_COLS

    y_all = np.concatenate([dataset.y_train, dataset.y_test])
    if colors is None:
        colors = {
            int(class_value): color
            for class_value, color in zip(
                np.unique(y_all), sns.color_palette("viridis", n_colors=len(np.unique(y_all)))
            )
        }

    if class_labels is None:
        class_labels = {class_value: str(class_value) for class_value in np.unique(y_all)}

    plt.figure(figsize=(10, 5), tight_layout=True)
    plt.subplot(1, 2, 1)
    for class_value in np.unique(dataset.y_train):
        subset = dataset.X_train[np.where(dataset.y_train == class_value)]
        plt.scatter(
            subset[:, 0],
            subset[:, 1],
            color=colors.get(class_value),
            label=class_labels.get(class_value),
        )
        plt.xlabel(f"{feature_cols[0]}")
        plt.ylabel(f"{feature_cols[1]}")
        plt.legend(title="Class")
        plt.title(str(kwargs.get("train_title", "Train Dataset")))

    plt.subplot(1, 2, 2)
    for class_value in np.unique(dataset.y_test):
        subset = dataset.X_test[np.where(dataset.y_test == class_value)]
        plt.scatter(
            subset[:, 0],
            subset[:, 1],
            color=colors.get(class_value),
            label=class_labels.get(class_value),
        )
        plt.xlabel(f"{feature_cols[0]}")
        plt.ylabel(f"{feature_cols[1]}")
        plt.legend(title="Class")
        plt.title(str(kwargs.get("test_title", "Test Dataset")))

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()



from qxmt.visualization.plot_classification_performance import (
    plot_2d_decision_boundaries,
    plot_2d_predicted_result,
)
from qxmt.visualization.plot_dataset import plot_2d_dataset
from qxmt.visualization.plot_metrics import plot_metric, plot_metrics_side_by_side
from qxmt.visualization.plot_regression_performance import (
    plot_actual_vs_predicted,
    plot_residual,
)

__all__ = [
    "plot_2d_dataset",
    "plot_metric",
    "plot_metrics_side_by_side",
    "plot_2d_decision_boundaries",
    "plot_2d_predicted_result",
    "plot_residual",
    "plot_actual_vs_predicted",
]



import subprocess
from logging import Logger
from pathlib import Path
from typing import Optional

from qxmt.constants import PROJECT_ROOT_DIR
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


def is_git_available() -> bool:
    """Check if git is available in the system.

    Returns:
        bool: whether git is available or not
    """
    try:
        result = subprocess.run(["git", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.returncode == 0
    except FileNotFoundError:
        return False


def get_commit_id(repo_path: Optional[Path | str] = None) -> str:
    """Get the commit ID of the current git repository.
    if the user is not using git version control, return "[Not using Git version control]".

    Args:
        repo_path (Optional[Path | str]): git repository path. Defaults to None.

    Returns:
        str: git commit ID
    """
    if repo_path is None:
        repo_path = PROJECT_ROOT_DIR

    command = ["git", "-C", str(repo_path), "rev-parse", "HEAD"]
    try:
        commit_id = subprocess.check_output(command).strip().decode("utf-8")
        return commit_id
    except subprocess.CalledProcessError:
        return "[Not using Git version control]"


def get_git_diff(repo_path: Optional[Path | str] = None, logger: Logger = LOGGER) -> str:
    """Get the git diff of the current git repository.
    if the current directory is not a git repository, return an empty string.

    Args:
        repo_path (Optional[Path | str]): git repository path. Defaults to None.
        logger (Logger): logger. Defaults to LOGGER.

    Returns:
        str: git diff
    """
    if repo_path is None:
        repo_path = PROJECT_ROOT_DIR

    command = ["git", "-C", str(repo_path), "diff"]
    try:
        diff = subprocess.check_output(command).strip().decode("utf-8")
        return diff
    except subprocess.CalledProcessError as e:
        command_str = " ".join(command)
        logger.warning(f'Error executing "{command_str}" command: {e}')
        return ""


def get_git_add_code(
    diff: Optional[str] = None,
    repo_path: Optional[Path | str] = None,
    logger: Logger = LOGGER,
) -> str:
    """Get the added code from the git diff.

    Args:
        diff (Optional[str]): string of git diff
        repo_path (Optional[Path | str]): git repository path. Defaults to None.
        logger (Logger): logger. Defaults to LOGGER.

    Returns:
        str: added code
    """
    if repo_path is None:
        repo_path = PROJECT_ROOT_DIR

    if diff is None:
        diff = get_git_diff(repo_path=repo_path, logger=logger)

    added_code = ""
    for line in diff.split("\n"):
        # ignore the line that starts with "+++" because it is the file
        if line.startswith("+") and not line.startswith("+++"):
            added_code += line[1:] + "\n"

    return added_code


def get_git_rm_code(
    diff: Optional[str] = None,
    repo_path: Optional[Path | str] = None,
    logger: Logger = LOGGER,
) -> str:
    """Get the removed code from the git diff.

    Args:
        diff (Optional[str]): string of git diff
        repo_path (Optional[Path | str]): git repository path. Defaults to None.
        logger (Logger): logger. Defaults to LOGGER.

    Returns:
        str: removed code
    """
    if repo_path is None:
        repo_path = PROJECT_ROOT_DIR

    if diff is None:
        diff = get_git_diff(repo_path=repo_path, logger=logger)

    removed_code = ""
    for line in diff.split("\n"):
        # ignore the line that starts with "---" because it is the file
        if line.startswith("-") and not line.startswith("---"):
            removed_code += line[1:] + "\n"

    return removed_code



class ExperimentSettingError(Exception):
    pass


class ExperimentRunSettingError(Exception):
    pass


class InvalidFileExtensionError(Exception):
    pass


class JsonEncodingError(Exception):
    pass


class ExperimentNotInitializedError(Exception):
    pass


class ReproductionError(Exception):
    pass


class InputShapeError(Exception):
    pass


class DeviceSettingError(Exception):
    pass


class ModelSettingError(Exception):
    pass


class InvalidPlatformError(Exception):
    pass


class InvalidQunatumDeviceError(Exception):
    pass


class InvalidModelNameError(Exception):
    pass


class InvalidConfigError(Exception):
    pass



from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Callable, Optional, cast

import numpy as np

from qxmt.constants import SUPPORTED_PLATFORMS
from qxmt.exceptions import InputShapeError, InvalidPlatformError
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class BaseFeatureMap(ABC):
    """Base class for feature map.
    This class is used to define the feature map for quantum machine learning.
    The feature map defined a quantum circuit that maps input data to quantum states.
    Provide a common interface within the QXMT library by absorbing differences between multiple platforms.
    User can define their own feature map by inheriting this class and defined in the 'feature_map' method.

    Examples:
        >>> import numpy as np
        >>> import pennylane as qml
        >>> from qxmt.feature_maps.base import BaseFeatureMap
        >>> class CustomFeatureMap(BaseFeatureMap):
        ...     def __init__(self, platform: str, n_qubits: int) -> None:
        ...         super().__init__(platform, n_qubits)
        ...
        ...     def feature_map(self, x: np.ndarray) -> None:
        ...         # define quantum circuit
        ...         qml.RX(x[0, 0], wires=0)
        ...
        >>> feature_map = CustomFeatureMap("pennylane", 2)
        >>> feature_map(np.random.rand(1, 2))
    """

    def __init__(self, platform: str, n_qubits: int) -> None:
        """Initialize the feature map class.
        [NOTE]: currently, only supports 'pennylane' platform. multi-platform support will be added in the future.

        Args:
            platform (str): name of quantum platform
            n_qubits (int): number of qubits
        """
        self.platform: str = platform
        self.n_qubits: int = n_qubits

        if self.platform not in SUPPORTED_PLATFORMS:
            raise InvalidPlatformError(
                f"Platform '{self.platform}' is not supported. Please choose in {SUPPORTED_PLATFORMS}."
            )

    def __call__(self, x: np.ndarray) -> None:
        self.feature_map(x)

    @abstractmethod
    def feature_map(self, x: np.ndarray) -> None:
        """Feature map function that maps input data to quantum states.

        Args:
            x (np.ndarray): input data
        """
        pass

    def check_input_dim_eq_nqubits(self, x: np.ndarray, idx: int = -1) -> None:
        """Check if the input data dimension matches the number of qubits.

        Args:
            x (np.ndarray): input data
            idx (int, optional): index of the dimension of qubit. Defaults to -1.

        Raises:
            InputShapeError: input data dimension does not match the number of qubits.
        """
        if x.shape[idx] != self.n_qubits:
            raise InputShapeError("Input data shape does not match the number of qubits.")

    def draw(
        self,
        x: Optional[np.ndarray] = None,
        x_dim: Optional[int] = None,
        format: str = "default",
        logger: Logger = LOGGER,
        **kwargs: Any,
    ) -> None:
        """Draw the circuit using the platform's draw function.

        Args:
            x (Optional[np.ndarray], optional): input example data for drawing the circuit. Defaults to None.
            x_dim (Optional[int], optional): dimension of input data. Defaults to None.
            format (str, optional): format of the drawing the circuit. Select "defalt" or "mpl". Defaults to "default".
            logger (Logger, optional): logger object. Defaults to LOGGER.

        Raises:
            NotImplementedError: not supported platform
        """
        if (x is None) and (x_dim is None):
            raise ValueError("Either 'x' or 'x_dim' argument must be provided.")

        x_sample = x[0] if x is not None else np.random.rand(1, cast(int, x_dim))[0]

        if self.platform == "pennylane":
            import pennylane as qml

            match format:
                case "default":
                    logger.info(qml.draw(qnode=self.feature_map, **kwargs)(x_sample))
                case "mpl":
                    logger.info(qml.draw_mpl(qnode=self.feature_map, **kwargs)(x_sample))
                case _:
                    raise ValueError(f"Invalid format '{format}' for drawing the circuit")
        else:
            raise NotImplementedError(f'"draw" method is not supported in {self.platform}.')


class FeatureMapFromFunc(BaseFeatureMap):
    """Wrap the feature map function to the BaseFeatureMap class."""

    def __init__(self, platform: str, n_qubits: int, feature_map_func: Callable[[np.ndarray], None]) -> None:
        super().__init__(platform, n_qubits)
        self.feature_map_func = feature_map_func

    def feature_map(self, x: np.ndarray) -> None:
        self.feature_map_func(x)



from typing import Optional

import numpy as np
from sklearn.preprocessing import StandardScaler

from qxmt.types import PROCESSCED_DATASET_TYPE


def normalization(
    X_train: np.ndarray,
    y_train: np.ndarray,
    X_val: Optional[np.ndarray],
    y_val: Optional[np.ndarray],
    X_test: np.ndarray,
    y_test: np.ndarray,
) -> PROCESSCED_DATASET_TYPE:
    """Normalization of dataset by StandardScaler

    Args:
        X_train (np.ndarray): numpy array of training data
        y_train (np.ndarray): numpy array of training label
        X_val (Optional[np.ndarray]): numpy array of validation data. None if validation set is not used
        y_val (Optional[np.ndarray]): numpy array of validation data. None if validation set is not used
        X_test (np.ndarray): numpy array of testing data
        y_test (np.ndarray): numpy array of testing label

    Returns:
        PROCESSCED_DATASET_TYPE: tuple of normalized dataset
    """
    scaler = StandardScaler()
    scaler.fit(X_train)
    X_train_scaled = scaler.transform(X_train)
    X_val_scaled = scaler.transform(X_val) if X_val is not None else None
    X_test_scaled = scaler.transform(X_test)

    return X_train_scaled, y_train, X_val_scaled, y_val, X_test_scaled, y_test



from qxmt.kernels.base import BaseKernel
from qxmt.kernels.sampling import (
    generate_all_observable_states,
    sample_results_to_probs,
    validate_sampling_values,
)

__all__ = ["BaseKernel", "generate_all_observable_states", "sample_results_to_probs", "validate_sampling_values"]



from qxmt.constants import PENNYLANE_DEVICES
from qxmt.devices.base import BaseDevice
from qxmt.exceptions import InvalidQunatumDeviceError
from qxmt.types import QuantumDeviceType


def get_platform_from_device(device: BaseDevice | QuantumDeviceType) -> str:
    """Get the platform name from the device.

    Args:
        device (BaseDevice | QuantumDeviceType): quantum device

    Returns:
        str: platform name
    """
    if isinstance(device, BaseDevice):
        return device.platform

    if isinstance(device, PENNYLANE_DEVICES):
        return "pennylane"
    else:
        raise InvalidQunatumDeviceError(f"Device {device} is not supported.")


def get_number_of_qubits(device: BaseDevice | QuantumDeviceType) -> int:
    """Get the number of qubits from the device.

    Args:
        device (BaseDevice | QuantumDeviceType): quantum device

    Returns:
        int: number of qubits
    """
    if isinstance(device, BaseDevice):
        return device.n_qubits

    if isinstance(device, PENNYLANE_DEVICES):
        return len(device.wires)
    else:
        raise InvalidQunatumDeviceError(f"Device {device} is not supported.")



import json
import os
import shutil
import time
from datetime import datetime
from logging import Logger
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd

from qxmt.configs import ExperimentConfig
from qxmt.constants import (
    DEFAULT_EXP_CONFIG_FILE,
    DEFAULT_EXP_DB_FILE,
    DEFAULT_EXP_DIRC,
    DEFAULT_MODEL_NAME,
    DEFAULT_N_JOBS,
    DEFAULT_SHOT_RESULTS_NAME,
    LLM_MODEL_PATH,
    TZ,
)
from qxmt.datasets.builder import DatasetBuilder
from qxmt.datasets.schema import Dataset
from qxmt.evaluation.evaluation import ClassificationEvaluation, RegressionEvaluation
from qxmt.exceptions import (
    ExperimentNotInitializedError,
    ExperimentRunSettingError,
    ExperimentSettingError,
    InvalidFileExtensionError,
    JsonEncodingError,
    ReproductionError,
)
from qxmt.experiment.schema import ExperimentDB, RunArtifact, RunRecord, RunTime
from qxmt.logger import set_default_logger
from qxmt.models.base import BaseMLModel
from qxmt.models.builder import ModelBuilder
from qxmt.utils import (
    get_commit_id,
    get_git_add_code,
    get_git_rm_code,
    is_git_available,
    save_experiment_config_to_yaml,
)

USE_LLM = os.getenv("USE_LLM", "FALSE").lower() == "true"
if USE_LLM:
    from qxmt.generators import DescriptionGenerator

IS_GIT_AVAILABLE = is_git_available()

LOGGER = set_default_logger(__name__)


class Experiment:
    """Experiment class for managing the experiment and each run data.
    The Experiment class provides methods for initializing the experiment,
    running the experiment, saving the experiment data, and reproducing the model.

    All experiment data is stored in the ExperimentDB instance.
    It is save in local directory as a json file (root_experiment_dirc/experiments/your_exp_name/experiment.json).

    Experiment can be initialized and strated from scratch by calling the `init()` method.
    Anthoer way is to load the existing experiment data from the json file (experiment.json)
    by calling the `load_experiment()` method.

    The Experiment class can be used in two ways:
    1. **Provide config_path**:
    This method accepts the path to the config file or config instance.
    It is more flexible but requires a YAML base config file.
    This method tracks the experiment settings, result and can reproduce the model.
    Officially, we recommend using the config file method.

    2. **Directly provide dataset and model instance**:
    This method directly accepts dataset and model instances.
    It is easy to use but does "NOT" track the experiment settings.
    This method is useful for adhoc experiments, quick testing or debugging.

    Examples:
        >>> import qxmt
        >>> exp = qxmt.Experiment(
        ...        name="my_qsvm_algorithm",
        ...       desc=\"""This is a experiment for new qsvm algorithm.
        ...        This experiment is applied and evaluated on multiple datasets.
        ...        \""",
        ...        auto_gen_mode=True,
        ...    ).init()
        >>> config_path = "../configs/template.yaml"
        >>> artifact, result = exp.run(
        ...     config_source=config_path)
        >>> exp.runs_to_dataframe()
            run_id	accuracy	precision	recall	f1_score
        0	     1	    0.45	     0.53	  0.66	    0.59
    """

    def __init__(
        self,
        name: Optional[str] = None,
        desc: Optional[str] = None,
        auto_gen_mode: bool = USE_LLM,
        root_experiment_dirc: str | Path = DEFAULT_EXP_DIRC,
        llm_model_path: str = LLM_MODEL_PATH,
        logger: Logger = LOGGER,
    ) -> None:
        """Initialize the Experiment class.
        Set the experiment name, description,
        and other settings such as auto_gen_mode, root_experiment_dirc and logger.
        auto_gen_mode controls whether to use the DescriptionGenerator by LLM.
        If use, set environemnt variable "USE_LLM" to True.
        root_experiment_dirc is the root directory to save the experiment data.
        Each artifact and result store in the subdirectory of the root directory.

        Args:
            name (Optional[str], optional):
                experiment name. If None, generate by execution time. Defaults to None.
            desc (Optional[str], optional):
                description of the experiment.
                The purpose is search, memo, etc not used in the code. Defaults to None.
            auto_gen_mode (bool, optional):
                whether to use the DescriptionGenerator
                for generating the description of each run. Defaults to USE_LLM.
            root_experiment_dirc (str | Path, optional):
                root directory to save the experiment data. Defaults to DEFAULT_EXP_DIRC.
            llm_model_path (str, optional):
                path to the LLM model. Defaults to LLM_MODEL_PATH.
            logger (Logger, optional):
                logger instance for warning or error messages. Defaults to LOGGER.


        """
        self.name: Optional[str] = name
        self.desc: Optional[str] = desc
        self.auto_gen_mode: bool = auto_gen_mode
        self.current_run_id: int = 0
        self.root_experiment_dirc: Path = Path(root_experiment_dirc)
        self.experiment_dirc: Path
        self.exp_db: Optional[ExperimentDB] = None
        self.logger: Logger = logger

        if (not USE_LLM) and (self.auto_gen_mode):
            self.logger.warning(
                'Global variable "USE_LLM" is set to False. '
                'DescriptionGenerator is not available. Set "USE_LLM" to True to use DescriptionGenerator.'
            )
            self.auto_gen_mode = False
        elif self.auto_gen_mode:
            self.desc_generator = DescriptionGenerator(llm_model_path)

    @staticmethod
    def _generate_default_name() -> str:
        """Generate a default name for the experiment.
        Default name is the current date and time in the format of
        "YYYYMMDDHHMMSSffffff"

        Returns:
            str: generated default name
        """
        return datetime.now(TZ).strftime("%Y%m%d%H%M%S%f")

    @staticmethod
    def _check_json_extension(file_path: str | Path) -> None:
        if Path(file_path).suffix.lower() != ".json":
            raise InvalidFileExtensionError(f"File '{file_path}' does not have a '.json' extension.")

    def init(self) -> "Experiment":
        """Initialize the experiment directory and DB.

        Returns:
            Experiment: initialized experiment
        """
        if self.name is None:
            self.name = self._generate_default_name()
        if self.desc is None:
            self.desc = ""

        self.experiment_dirc = self.root_experiment_dirc / self.name
        # create experiment directory if it does not exist or empty
        if (not self.experiment_dirc.exists()) or (not any(self.experiment_dirc.iterdir())):
            self.experiment_dirc.mkdir(parents=True, exist_ok=True)
        else:
            raise ExperimentSettingError(f"Experiment directory '{self.experiment_dirc}' already exists.")

        self.exp_db = ExperimentDB(
            name=self.name,
            desc=self.desc,
            working_dirc=Path.cwd(),
            experiment_dirc=self.experiment_dirc,
            runs=[],
        )

        return self

    def load(self, exp_dirc: str | Path, exp_file_name: str | Path = DEFAULT_EXP_DB_FILE) -> "Experiment":
        """Load existing experiment data from a json file.

        Args:
            exp_dirc (str | Path): path to the experiment directory

        Raises:
            FileNotFoundError: if the experiment file does not exist
            ExperimentSettingError: if the experiment directory does not exist

        Returns:
            Experiment: loaded experiment
        """
        exp_file_path = Path(exp_dirc) / exp_file_name
        if not exp_file_path.exists():
            raise FileNotFoundError(f"{exp_file_path} does not exist.")

        self._check_json_extension(exp_file_path)
        with open(exp_file_path, "r") as json_file:
            exp_data = json.load(json_file)

        # set the experiment data from the json file
        self.exp_db = ExperimentDB(**exp_data)

        # update the experiment name, description, working directory, and experiment directory
        # if the loaded data is different from the current settings
        if (self.name is not None) and (self.name != self.exp_db.name):
            self.exp_db.name = self.name
            self.logger.info(f'Name is changed from "{self.exp_db.name}" to "{self.name}".')
        else:
            self.name = self.exp_db.name

        if (self.desc is not None) and (self.desc != self.exp_db.desc):
            self.exp_db.desc = self.desc
            self.logger.info(f'Description is changed from "{self.exp_db.desc}" to "{self.desc}".')
        else:
            self.desc = self.exp_db.desc

        working_dirc = Path.cwd()
        if working_dirc != self.exp_db.working_dirc:
            self.logger.info(f'Working directory is changed from "{self.exp_db.working_dirc}" to "{working_dirc}".')
            self.exp_db.working_dirc = working_dirc

        self.experiment_dirc = self.root_experiment_dirc / str(self.name)
        if self.experiment_dirc != self.exp_db.experiment_dirc:
            self.logger.info(
                f'Experiment directory is changed from "{self.exp_db.experiment_dirc}" to "{self.experiment_dirc}".'
            )
            self.exp_db.experiment_dirc = self.experiment_dirc

        if not self.exp_db.experiment_dirc.exists():
            raise ExperimentSettingError(f"Experiment directory '{self.exp_db.experiment_dirc}' does not exist.")

        self.current_run_id = len(self.exp_db.runs)
        self.save_experiment()

        return self

    def _is_initialized(self) -> None:
        """Check if the experiment is initialized.

        Raises:
            ExperimentNotInitializedError: if the experiment is not initialized
        """
        if self.exp_db is None:
            raise ExperimentNotInitializedError(
                "Experiment is not initialized. Please call init() or load_experiment() method first."
            )

    def _run_setup(self) -> Path:
        """Setup for the current run.
        Create a new run directory and update the current run ID.
        If the run directory already exists, raise an error and not update the current run ID.

        Returns:
            Path: path to the current run directory

        Raises:
            Exception: if error occurs while creating the run directory
        """
        current_run_id = self.current_run_id + 1
        try:
            current_run_dirc = self.experiment_dirc / f"run_{current_run_id}"
            current_run_dirc.mkdir(parents=True)
        except Exception as e:
            self.logger.error(f"Error creating run directory: {e}")
            raise

        self.current_run_id = current_run_id

        return current_run_dirc

    def _run_backfill(self) -> None:
        """Backfill the current run when an error occurs during the run."""
        run_dirc = self.experiment_dirc / f"run_{self.current_run_id}"
        if run_dirc.exists():
            shutil.rmtree(run_dirc)

        self.current_run_id -= 1

    def run_evaluation(
        self,
        task_type: str,
        actual: np.ndarray,
        predicted: np.ndarray,
        default_metrics_name: Optional[list[str]],
        custom_metrics: Optional[list[dict[str, Any]]],
    ) -> dict:
        """Run evaluation for the current run.

        Args:
            actual (np.ndarray): array of actual values
            predicted (np.ndarray): array of predicted values
            default_metrics_name (Optional[list[str]]): list of default metrics name
            custom_metrics (Optional[list[dict[str, Any]]]): list of user defined custom metric configurations

        Returns:
            dict: evaluation result
        """
        match task_type:
            case "classification":
                evaluation = ClassificationEvaluation(
                    actual=actual,
                    predicted=predicted,
                    default_metrics_name=default_metrics_name,
                    custom_metrics=custom_metrics,
                )
            case "regression":
                evaluation = RegressionEvaluation(
                    actual=actual,
                    predicted=predicted,
                    default_metrics_name=default_metrics_name,
                    custom_metrics=custom_metrics,
                )
            case _:
                raise ValueError(f"Invalid task_type: {task_type}")

        evaluation.evaluate()

        return evaluation.to_dict()

    def _run_from_config(
        self,
        config: ExperimentConfig,
        commit_id: str,
        run_dirc: str | Path,
        n_jobs: int,
        repo_path: Optional[str] = None,
        add_results: bool = True,
    ) -> tuple[RunArtifact, RunRecord]:
        """Run the experiment from the config file.

        Args:
            config (ExperimentConfig): configuration of the experiment
            commit_id (str): commit ID of the current git repository
            run_dirc (str | Path): path to the run directory
            n_jobs (int): number of jobs for parallel processing
            repo_path (str, optional): path to the git repository. Defaults to None.
            add_results (bool, optional): whether to save the model. Defaults to True.

        Returns:
            tuple[RunArtifact, RunRecord]: artifact and record of the current run_id
        """
        # create dataset instance from pre defined raw_preprocess_logic and transform_logic
        dataset = DatasetBuilder(config=config).build()

        # create model instance from the config
        model = ModelBuilder(config=config, n_jobs=n_jobs).build()
        save_shots_path = Path(run_dirc) / DEFAULT_SHOT_RESULTS_NAME if add_results else None
        save_model_path = Path(run_dirc) / DEFAULT_MODEL_NAME

        artifact, record = self._run_from_instance(
            task_type=config.global_settings.task_type,
            dataset=dataset,
            model=model,
            save_shots_path=save_shots_path,
            save_model_path=save_model_path,
            default_metrics_name=config.evaluation.default_metrics,
            custom_metrics=config.evaluation.custom_metrics,
            desc=config.description,
            commit_id=commit_id,
            config_file_name=DEFAULT_EXP_CONFIG_FILE,
            repo_path=repo_path,
            add_results=add_results,
        )

        return artifact, record

    def _run_from_instance(
        self,
        task_type: str,
        dataset: Dataset,
        model: BaseMLModel,
        save_shots_path: Optional[str | Path],
        save_model_path: str | Path,
        default_metrics_name: Optional[list[str]],
        custom_metrics: Optional[list[dict[str, Any]]],
        desc: str,
        commit_id: str,
        config_file_name: Path,
        repo_path: Optional[str] = None,
        add_results: bool = True,
    ) -> tuple[RunArtifact, RunRecord]:
        """Run the experiment from the dataset and model instance.

        Args:
            task_type (str): type of the task (classification or regression)
            dataset (Dataset): dataset object
            model (BaseMLModel): model object
            save_model_path (str | Path): path to save the model
            default_metrics_name (Optional[list[str]]): list of default metrics name
            custom_metrics (Optional[list[dict[str, Any]]]): list of user defined custom metric configurations
            desc (str, optional): description of the run.
            commit_id (str): commit ID of the current git repository
            config_file_name (Path): name of the config file
            repo_path (str, optional): path to the git repository. Defaults to None.
            add_results (bool, optional): whether to save the model. Defaults to True.

        Returns:
            tuple[RunArtifact, RunRecord]: artifact and record of the current run_id
        """
        fit_start = time.perf_counter()
        model.fit(X=dataset.X_train, y=dataset.y_train, save_shots_path=save_shots_path)
        fit_end = time.perf_counter()

        predict_start = time.perf_counter()
        predicted = model.predict(dataset.X_test)
        predict_end = time.perf_counter()

        if add_results:
            model.save(save_model_path)

        if self.auto_gen_mode and (desc == ""):
            if IS_GIT_AVAILABLE:
                desc = self.desc_generator.generate(
                    add_code=get_git_add_code(repo_path=repo_path, logger=self.logger),
                    remove_code=get_git_rm_code(repo_path=repo_path, logger=self.logger),
                )
            else:
                self.logger.warning(
                    """Git command is not available. DescriptionGenerator need git environment to generate.
                        Current run description set to empty string."""
                )
                desc = ""

        artifact = RunArtifact(
            run_id=self.current_run_id,
            dataset=dataset,
            model=model,
        )

        record = RunRecord(
            run_id=self.current_run_id,
            desc=desc,
            commit_id=commit_id,
            config_file_name=config_file_name,
            execution_time=datetime.now(TZ).strftime("%Y-%m-%d %H:%M:%S.%f %Z%z"),
            runtime=RunTime(fit_seconds=fit_end - fit_start, predict_seconds=predict_end - predict_start),
            evaluation=self.run_evaluation(
                task_type=task_type,
                actual=dataset.y_test,
                predicted=predicted,
                default_metrics_name=default_metrics_name,
                custom_metrics=custom_metrics,
            ),
        )

        return artifact, record

    def run(
        self,
        task_type: Optional[str] = None,
        dataset: Optional[Dataset] = None,
        model: Optional[BaseMLModel] = None,
        config_source: Optional[ExperimentConfig | str | Path] = None,
        default_metrics_name: Optional[list[str]] = None,
        custom_metrics: Optional[list[dict[str, Any]]] = None,
        n_jobs: int = DEFAULT_N_JOBS,
        desc: str = "",
        repo_path: Optional[str] = None,
        add_results: bool = True,
    ) -> tuple[RunArtifact, RunRecord]:
        """
        Start a new run for the experiment.

        The `run()` method can be called in two ways:

        1. **Provide dataset and model instance**:
        This method directly accepts dataset and model instances.
        It is easy to use but less flexible and does "NOT" track the experiment settings.

        2. **Provide config_path**:
        This method accepts the path to the config file or config instance.
        It is more flexible but requires a config file.

        Args:
            task_type (str, optional): type of the task (classification or regression). Defaults to None.
            dataset (Dataset): the dataset object.
            model (BaseMLModel): the model object.
            config_source (ExperimentConfig, str | Path, optional): config source can be either an `ExperimentConfig`
                instance or the path to a config file. If a path is provided, it loads and creates an
                `ExperimentConfig` instance. Defaults to None.
            default_metrics_name (list[str], optional): list of default metrics names. Defaults to None.
            custom_metrics (list[dict[str, Any]], optional):
                list of user defined custom metric configurations. Defaults to None.
            n_jobs (int, optional): number of jobs for parallel processing. Defaults to DEFAULT_N_JOBS.
            desc (str, optional): description of the run. Defaults to "".
            repo_path (str, optional): path to the git repository. Defaults to None.
            add_results (bool, optional): whether to add the run record to the experiment. Defaults to True.

        Returns:
            tuple[RunArtifact, RunRecord]: Returns a tuple containing the artifact and run record of the current run_id.

        Raises:
            ExperimentNotInitializedError: Raised if the experiment is not initialized.
        """
        self._is_initialized()

        if add_results:
            current_run_dirc = self._run_setup()
            commit_id = get_commit_id(repo_path=repo_path) if IS_GIT_AVAILABLE else ""
        else:
            current_run_dirc = Path("")
            commit_id = ""

        try:
            if config_source is not None:
                if isinstance(config_source, str | Path):
                    config = ExperimentConfig(path=config_source)
                else:
                    config = config_source

                artifact, record = self._run_from_config(
                    config=config,
                    commit_id=commit_id,
                    run_dirc=current_run_dirc,
                    n_jobs=n_jobs,
                    repo_path=repo_path,
                    add_results=add_results,
                )
            elif (dataset is not None) and (model is not None):
                if task_type is None:
                    raise ExperimentRunSettingError(
                        """
                        task_type must be provided when dataset and model are provided.
                        Please provide task_type="classification" or "regression".
                        """
                    )
                artifact, record = self._run_from_instance(
                    task_type=task_type,
                    dataset=dataset,
                    model=model,
                    save_shots_path=current_run_dirc / DEFAULT_SHOT_RESULTS_NAME if add_results else None,
                    save_model_path=current_run_dirc / DEFAULT_MODEL_NAME,
                    default_metrics_name=default_metrics_name,
                    custom_metrics=custom_metrics,
                    desc=desc,
                    commit_id=commit_id,
                    config_file_name=Path(""),
                    repo_path=repo_path,
                    add_results=add_results,
                )
            else:
                raise ExperimentRunSettingError("Either dataset and model or config must be provided.")
        except Exception as e:
            self.logger.error(f"Error occurred during the run: {e}")
            if add_results:
                self._run_backfill()
            raise e

        if add_results:
            self.exp_db.runs.append(record)  # type: ignore
            self.save_experiment()

            if config_source is not None:
                # config is saved in the run directory
                # if "run" executed by instance mode, config is not saved
                save_experiment_config_to_yaml(
                    config, Path(current_run_dirc) / DEFAULT_EXP_CONFIG_FILE, delete_source_path=True
                )

        return artifact, record

    def runs_to_dataframe(self) -> pd.DataFrame:
        """Convert the run data to a pandas DataFrame.

        Returns:
            pd.DataFrame: DataFrame of run data

        Raises:
            ExperimentNotInitializedError: if the experiment is not initialized
        """
        self._is_initialized()
        run_data = [run.model_dump() for run in self.exp_db.runs]  # type: ignore
        run_data = [
            {"run_id": run_record_dict["run_id"], **run_record_dict["evaluation"]} for run_record_dict in run_data
        ]
        return pd.DataFrame(run_data)

    def save_experiment(self, exp_file: str | Path = DEFAULT_EXP_DB_FILE) -> None:
        """Save the experiment data to a json file.

        Args:
            exp_file (str | Path, optional):
                name of the file to save the experiment data.Defaults to DEFAULT_EXP_DB_FILE.

        Raises:
            ExperimentNotInitializedError: if the experiment is not initialized
        """

        def custom_encoder(obj: Any) -> str:
            if isinstance(obj, Path):
                return str(obj)
            raise JsonEncodingError(f"Object of type {type(obj).__name__} is not JSON serializable")

        self._is_initialized()
        save_path = self.experiment_dirc / exp_file
        self._check_json_extension(save_path)
        exp_data = json.loads(self.exp_db.model_dump_json())  # type: ignore
        with open(save_path, "w") as json_file:
            json.dump(exp_data, json_file, indent=4, default=custom_encoder)

    def get_run_record(self, runs: list[RunRecord], run_id: int) -> RunRecord:
        """Get the run record of the target run_id.

        Args:
            run_id (int): target run_id

        Raises:
            ValueError: if the run record does not exist

        Returns:
            RunRecord: target run record
        """
        self._is_initialized()
        for run_record in runs:
            if run_record.run_id == run_id:
                return run_record

        # if the target run_id does not exist
        raise ValueError(f"Run record of run_id={run_id} does not exist.")

    def _validate_evaluation(
        self, logging_evaluation: dict[str, float], reproduction_evaluation: dict[str, float]
    ) -> None:
        """Validate the evaluation results of logging and reproduction.

        Args:
            logging_evaluation (dict[str, float]): evaluation result of logging
            reproduction_evaluation (dict[str, float]): evaluation result of reproduction

        Raises:
            ValueError: if the evaluation results are different
        """
        invalid_dict: dict[str, str] = {}
        for key, value in logging_evaluation.items():
            reproduction_value = reproduction_evaluation.get(key, None)
            if value != reproduction_value:
                invalid_dict[key] = f"{value} -> {reproduction_value}"

        if len(invalid_dict) > 0:
            raise ValueError(
                f"Evaluation results are different between logging and reproduction (invalid metrics: {invalid_dict})."
            )

    def reproduce(self, run_id: int, check_commit_id: bool = False) -> tuple[RunArtifact, RunRecord]:
        """Reproduce the target run_id model from config file.
        If the target run_id does not have a config file path, raise an error.
        Reoroduce method not supported for the run executed from the instance.

        Args:
            run_id (int): target run_id
            check_commit_id (bool, optional): whether to check the commit_id. Defaults to False.

        Returns:
            tuple[RunArtifact, RunRecord]: artifact and record of the reproduced run_id

        Raises:
            ReproductinoError: if the run_id does not have a config file path
        """
        self._is_initialized()
        run_record = self.get_run_record(self.exp_db.runs, run_id)  # type: ignore

        if check_commit_id:
            commit_id = get_commit_id() if IS_GIT_AVAILABLE else ""
            if commit_id != run_record.commit_id:
                self.logger.warning(
                    f'Current commit_id="{commit_id}" is different from'
                    f'the run_id={run_id} commit_id="{run_record.commit_id}".'
                )
        if run_record.config_file_name == Path(""):
            raise ReproductionError(
                f"run_id={run_id} does not have a config file path. This run executed from instance."
                "run from instance mode not supported for reproduction."
            )

        config_path = Path(f"{self.experiment_dirc}/run_{run_id}/{DEFAULT_EXP_CONFIG_FILE}")
        reproduced_artifact, reproduced_result = self.run(config_source=config_path, add_results=False)

        logging_evaluation = run_record.evaluation
        reproduced_evaluation = reproduced_result.evaluation
        self._validate_evaluation(logging_evaluation, reproduced_evaluation)
        self.logger.info(f"Reproduce model is successful. Evaluation results are the same as run_id={run_id}.")

        return reproduced_artifact, reproduced_result



from logging import Logger
from pathlib import Path
from typing import Optional, cast

import numpy as np
import openml
import pandas as pd

from qxmt.decorators import retry_on_exception
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class OpenMLDataLoader:
    """
    This class loads a dataset from OpenML (https://www.openml.org/) and converts it to the specified return format.
    OpenML is an online platform for sharing datasets and machine learning experiments.
    Identifier of the dataset can be specified by the dataset name or ID.
    If the name is specified, search the OpenML database to get the dataset ID and load the latest version.
    The dataset ID can be search in officail OpenML website (https://www.openml.org/search).
    ID has a higher priority than the name.

    Supported return formats:
    - numpy: return as a tuple of numpy arrays.
    - pandas: return as a pandas DataFrame.

    Supported save formats:
    - numpy: .npz, .npy
    - pandas: .csv, .tsv

    Examples:

        Load the dataset "mnist_784" from OpenML and save it as a numpy file.

        ::

            >>> loader = OpenMLDataLoader(name="mnist_784", save_path="data/mnist.npz")
            >>> X, y = loader.load()
            X = np.array([[0, 0, 0, ..., 0, 0, 0], ..]), y = np.array([5, 0, 4, ..., 4, 5, 6])

        Load the dataset id=554 ("mnist_784") from OpenML and save it as a pandas DataFrame.

        ::

            >>> loader = OpenMLDataLoader(id=554, save_path="data/mnist.csv", return_format="pandas")
            >>> data = loader.load()
            data = pd.DataFrame([[0, 0, 0, ..., 0, 0, 0], .., [5, 0, 4, ..., 4, 5, 6]])
    """

    def __init__(
        self,
        name: Optional[str] = None,
        id: Optional[int] = None,
        save_path: Optional[str | Path] = None,
        return_format: str = "numpy",
        use_cache: bool = True,
        logger: Logger = LOGGER,
    ) -> None:
        """Initialize the OpenML dataset loader.

        Args:
            dataset_identifier (str | int): dataset name or ID. If the name is specified, the latest version is used.
            save_path (Optional[str  |  Path], optional):
                save path for the loaded dataset.
                If the value is None, the dataset is not saved. Defaults to None.
            return_format (str, optional): return format of the loaded dataset. Defaults to "numpy".
        """
        if (name is None) and (id is None):
            raise ValueError("Either dataset 'name' or 'id' must be specified.")
        self.name = name
        self.id = id if id is not None else self._get_dataset_id()
        self.save_path = save_path
        self.return_format = return_format.lower()
        self.use_cache = use_cache
        self.logger = logger

    @retry_on_exception(retries=3, delay=5)
    def _get_dataset_id(self) -> int:
        """Get the dataset ID from the dataset name by searching the OpenML database.

        Raises:
            ValueError: dataset that matches the name is not found in the OpenML database.

        Returns:
            int: mathced dataset ID
        """
        datasets = openml.datasets.list_datasets(output_format="dataframe")
        matches = datasets[datasets["name"] == self.name]
        if not matches.empty:
            # get the dataset id of the latest version
            matches = matches.sort_values("version", ascending=False)
            return int(matches.iloc[0]["did"])
        else:
            raise ValueError(f"Dataset '{self.name}' not found on OpenML.")

    def load(self) -> tuple[np.ndarray, np.ndarray | None] | pd.DataFrame:
        """Load the dataset from OpenML. Then, convert the dataset to the specified return format.
        The loaded dataset is saved to the specified path if it is not None.

        Raises:
            ValueError: unsupported return format

        Returns:
            tuple[np.ndarray, np.ndarray | None] | pd.DataFrame: loaded dataset
        """
        dataset = openml.datasets.get_dataset(
            self.id,
            download_data=False,
            download_qualities=False,
            download_features_meta_data=False,
            force_refresh_cache=not self.use_cache,
        )
        self.logger.debug(f"Loading dataset: {dataset.name}, ID: {self.id}")

        X, y, categorical_indicator, attribute_names = dataset.get_data(
            dataset_format="dataframe", target=dataset.default_target_attribute
        )

        data = self._format_return_data(cast(pd.DataFrame, X), cast(pd.DataFrame, y))

        if self.save_path:
            self._save_dataset(data)

        return data

    def _format_return_data(
        self, X: pd.DataFrame, y: pd.DataFrame | None
    ) -> tuple[np.ndarray, np.ndarray | None] | pd.DataFrame:
        """Format the loaded dataset to the specified return format.
        Return format defined in the class attribute `return_format`.
        supported formats:
        - numpy: return as a tuple of numpy arrays.
        - pandas: return as a pandas DataFrame.

        Args:
            X (pd.DataFrame): DataFrame of features.
            y (pd.DataFrame | None): DataFrame of labels.

        Raises:
            ValueError: unsupported return format

        Returns:
            tuple[np.ndarray, np.ndarray | None] | pd.DataFrame: formated dataset for the specified return format.
        """
        if self.return_format in ["numpy", "array"]:
            X_array = X.to_numpy()
            y_array = y.to_numpy() if y is not None else None
            data = (X_array, y_array)
        elif self.return_format in ["pandas", "dataframe"]:
            if y is None:
                data = X
            else:
                data = pd.concat([X, y], axis=1)
        else:
            raise ValueError(f"Unsupported return format: {self.return_format}")

        return data

    def _save_dataset(self, data: tuple[np.ndarray, np.ndarray | None] | pd.DataFrame) -> None:
        """Save the loaded dataset to the specified path.
        Supported save formats:
        - numpy: .npz, .npy
        - pandas: .csv, .tsv

        Args:
            data (tuple[np.ndarray, np.ndarray  |  None] | pd.DataFrame): dataset object to save.

        Raises:
            ValueError: save path is not specified.
            ValueError: not supported save format
        """
        if self.save_path is None:
            raise ValueError("Save path is not specified.")

        save_path = Path(self.save_path)
        # create the directory if it does not exist
        save_path.parent.mkdir(parents=True, exist_ok=True)

        extention = save_path.suffix
        if self.return_format in ["numpy", "array"] and extention == ".npz":
            if data[1] is not None:
                np.savez(save_path, X=cast(np.ndarray, data[0]), y=cast(np.ndarray, data[1]))
            else:
                np.savez(save_path, X=cast(np.ndarray, data[0]))
        elif self.return_format in ["numpy", "array"] and extention == ".npy":
            if data[1] is not None:
                np.save(save_path.with_name(save_path.stem + "_X" + save_path.suffix), cast(np.ndarray, data[0]))
                np.save(save_path.with_name(save_path.stem + "_y" + save_path.suffix), cast(np.ndarray, data[1]))
            else:
                np.save(save_path, cast(np.ndarray, data[0]))
        elif self.return_format in ["pandas", "dataframe"] and extention == ".csv":
            pd.DataFrame(data).to_csv(save_path, sep=",", index=False)
        elif self.return_format in ["pandas", "dataframe"] and extention == ".tsv":
            pd.DataFrame(data).to_csv(save_path, sep="\t", index=False)
        else:
            raise ValueError(f'Unsupported save format: data_format="{self.return_format}", extention="{extention}"')

        self.logger.debug(f"Saved dataset to: {save_path.parent}")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import numpy as np

from qxmt.constants import DEFAULT_N_JOBS
from qxmt.feature_maps.base import BaseFeatureMap
from qxmt.kernels.base import BaseKernel


class BaseMLModel(ABC):
    """Base class for quantum machine learning models.
    This class is an abstract class for qunatum machine learning models.
    If you want to implement a new quantum machine learning model, you should inherit this class.
    For compatibility with QXMT framework, you should implement some methods such as fit,
    predict, save, load, get_params, and set_params.
    """

    @abstractmethod
    def fit(self, X: np.ndarray, y: np.ndarray, **kwargs: Any) -> None:
        """Fit the model with given data.

        Args:
            X (np.ndarray): array of features
            y (np.ndarray): array of target value
        """
        pass

    @abstractmethod
    def predict(self, X: np.ndarray) -> np.ndarray:
        """Predict the target value with given features.

        Args:
            X (np.ndarray): array of features

        Returns:
            np.ndarray: array of predicted value
        """
        pass

    @abstractmethod
    def save(self, path: str | Path) -> None:
        """Save the model to the given path.
        [TODO]: standardize the save/load method using same library

        Args:
            path (str | Path): path to save the model
        """
        pass

    @abstractmethod
    def load(self, path: str | Path) -> "BaseMLModel":
        """Load the model from the given path.
        [TODO]: standardize the save/load method using same library

        Args:
            path (str | Path): path to load the model
        """
        pass

    @abstractmethod
    def get_params(self) -> dict:
        """Get the parameters of the model.

        Returns:
            dict: dictionary of parameters
        """
        pass

    @abstractmethod
    def set_params(self, params: dict) -> None:
        """Set the parameters of the model.

        Args:
            params (dict): dictionary of parameters
        """
        pass


class BaseKernelModel(BaseMLModel):
    """Base class for kernel-based quantum machine learning models.
    This class is an abstract class for kernel-based quantum machine learning models.
    If you want to implement a new kernel-based quantum machine learning model, you should inherit this class.
    This class requires a kernel instance of BaseKernel class.
    The user can use any Feature Map or Kernel to be used, as long as it follows the interface of the BaseKernel class.
    """

    def __init__(self, kernel: BaseKernel, n_jobs: int = DEFAULT_N_JOBS) -> None:
        """Initialize the kernel model.

        Args:
            kernel (BaseKernel): kernel instance of BaseKernel class
        """
        super().__init__()
        self.kernel = kernel
        self.n_jobs = n_jobs

    def get_feature_map(self) -> BaseFeatureMap:
        """Get the feature map of the model.

        Returns:
            BaseFeatureMap: feature map instance
        """
        return self.kernel.feature_map

    def get_kernel_matrix(
        self,
        x_array_1: np.ndarray,
        x_array_2: np.ndarray,
    ) -> np.ndarray:
        """Get the kernel matrix of the given data.
        This method is alias of kernel.compute_matrix().

        Args:
            x_array_1 (np.ndarray): array of samples (ex: training data)
            x_array_2 (np.ndarray): array of samples (ex: test data)

        Returns:
            np.ndarray: kernel matrix
        """
        kernel_matrix, _ = self.kernel.compute_matrix(
            x_array_1,
            x_array_2,
            return_shots_resutls=False,
            n_jobs=self.n_jobs,
        )

        return kernel_matrix

    def plot_kernel_matrix(
        self,
        x_array_1: np.ndarray,
        x_array_2: np.ndarray,
    ) -> None:
        """Plot the kernel matrix of the given data.
        This method is alias of kernel.plot_matrix().

        Args:
            x_array_1 (np.ndarray): array of samples (ex: training data)
            x_array_2 (np.ndarray): array of samples (ex: test data)
        """
        self.kernel.plot_matrix(x_array_1, x_array_2, n_jobs=self.n_jobs)

    def plot_train_test_kernel_matrix(
        self,
        x_train: np.ndarray,
        x_test: np.ndarray,
    ) -> None:
        """Plot the kernel matrix of training and testing data.
        This method is alias of kernel.plot_train_test_matrix().

        Args:
            x_train (np.ndarray): array of training samples
            x_test (np.ndarray): array of testing samples
        """
        self.kernel.plot_train_test_matrix(x_train, x_test, n_jobs=self.n_jobs)



import numpy as np
import pennylane as qml

from qxmt.constants import PENNYLANE_PLATFORM
from qxmt.feature_maps import BaseFeatureMap


class NPQCFeatureMap(BaseFeatureMap):
    """NPQC feature map class.
    Reference: https://arxiv.org/abs/2108.01039

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.npqc import NPQCFeatureMap
        >>> feature_map = NPQCFeatureMap(2, 2, 0.1)
        >>> feature_map.draw(x_dim=2)
        0: ──RY(1.60)──RZ(1.59)──RY(1.57)─╭●──RY(1.60)──RZ(1.59)──RY(1.57)─╭●──RY(1.60)─┤
        1: ──RY(1.60)──RZ(1.59)───────────╰Z───────────────────────────────╰Z───────────┤
    """

    def __init__(self, n_qubits: int, reps: int, c: float) -> None:
        """ "Initialize the NPQC feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
            c (float): scaling factor
            seed (int): random seed
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.n_qubits: int = n_qubits
        self.reps: int = reps
        self.c: float = c
        self._validation()

    def _validation(self) -> None:
        """Validate the NPQC feature map."""
        if self.n_qubits % 2 != 0:
            raise ValueError(f"NPQC feature map requires an even number of qubits. but got {self.n_qubits}")

    def _calculate_target_wire(self, idx: int, r_idx: int, n_qubits: int) -> int:
        """Calculate target wire for controlled-Z gate.

        Args:
            idx (int): source wire index
            r_idx (int): repetition index
            n_qubits (int): number of qubits

        Returns:
            int: target wire index
        """
        remaining_divisions = r_idx + 1
        target_offset = 0
        while remaining_divisions % 2 == 0:
            remaining_divisions //= 2
            target_offset += 1
        return (idx + target_offset * 2 + 1) % n_qubits

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of NPQC feature map.

        Args:
            x (np.ndarray): input data
        """
        data_idx = 0
        # Apply RY and RZ rotations based on input
        for i in range(self.n_qubits):
            qml.RY(self.c * x[data_idx % len(x)] + np.pi / 2, wires=i)
            data_idx += 1
            qml.RZ(self.c * x[data_idx % len(x)] + np.pi / 2, wires=i)
            data_idx += 1

        for r_idx in range(self.reps):
            for i in range(0, self.n_qubits - 1, 2):
                qml.RY(np.pi / 2, wires=i)

                # Calculate and apply controlled-Z gate
                target_wire = self._calculate_target_wire(i, r_idx, self.n_qubits)
                qml.CZ(wires=[i, target_wire])

                # Add RY and optional RZ gates for parameterized input
                qml.RY(self.c * x[data_idx % len(x)] + np.pi / 2, wires=i)
                data_idx += 1
                if r_idx + 1 < self.reps:
                    qml.RZ(self.c * x[data_idx % len(x)] + np.pi / 2, wires=i)
                    data_idx += 1



from typing import Callable, cast

import numpy as np
import pennylane as qml
from pennylane.measurements.probs import ProbabilityMP
from pennylane.measurements.sample import SampleMP

from qxmt.devices.base import BaseDevice
from qxmt.exceptions import ModelSettingError
from qxmt.feature_maps.base import BaseFeatureMap
from qxmt.kernels.base import BaseKernel
from qxmt.kernels.sampling import sample_results_to_probs


class FidelityKernel(BaseKernel):
    """Fidelity kernel class.
    The fidelity kernel is a quantum kernel that computes the kernel value based on the fidelity
    between two quantum states.

    Args:
        BaseKernel (_type_): base class of kernel

    Examples:
        >>> import numpy as np
        >>> from qxmt.kernels.pennylane.fidelity_kernel import FidelityKernel
        >>> from qxmt.feature_maps.pennylane.defaults import ZZFeatureMap
        >>> from qxmt.configs import DeviceConfig
        >>> from qxmt.devices.builder import DeviceBuilder
        >>> config = DeviceConfig(
        ...     platform="pennylane",
        ...     name="default.qubit",
        ...     n_qubits=2,
        ...     shots=1024,
        >>> )
        >>> device = DeviceBuilder(config).build()
        >>> feature_map = ZZFeatureMap(2, 2)
        >>> kernel = FidelityKernel(device, feature_map)
        >>> x1 = np.random.rand(2)
        >>> x2 = np.random.rand(2)
        >>> kernel.compute(x1, x2)
        0.14
    """

    def __init__(
        self,
        device: BaseDevice,
        feature_map: BaseFeatureMap | Callable[[np.ndarray], None],
    ) -> None:
        """Initialize the FidelityKernel class.

        Args:
            device (BaseDevice): device instance for quantum computation
            feature_map (BaseFeatureMap | Callable[[np.ndarray], None]): feature map instance or function
        """

        super().__init__(device, feature_map)

    def _circuit(self, x1: np.ndarray, x2: np.ndarray) -> ProbabilityMP | SampleMP:
        if self.feature_map is None:
            raise ModelSettingError("Feature map must be provided for FidelityKernel.")

        self.feature_map(x1)
        qml.adjoint(self.feature_map)(x2)  # type: ignore

        if self.is_sampling:
            return qml.sample(wires=range(self.n_qubits))
        else:
            return qml.probs(wires=range(self.n_qubits))

    def compute(self, x1: np.ndarray, x2: np.ndarray) -> tuple[float, np.ndarray]:
        """Compute the fidelity kernel value between two data points.

        Args:
            x1 (np.ndarray): numpy array representing the first data point
            x2 (np.ndarray): numpy array representing the second data point

        Returns:
            tuple[float, np.ndarray]: fidelity kernel value and probability distribution
        """
        qnode = qml.QNode(self._circuit, device=self.device.get_simulator(), cache=False)  # type: ignore
        result = qnode(x1, x2)

        if self.is_sampling:
            # convert the sample results to probability distribution
            # shots must be over 0 when sampling mode
            probs = sample_results_to_probs(result, self.n_qubits, cast(int, self.device.shots))
        else:
            # use theoretical probability distribution
            probs = np.array(result)

        kernel_value = probs[0]  # get |0..0> state probability

        return kernel_value, probs



import numpy as np

from qxmt.types import RAW_DATASET_TYPE


def sampling_by_num(X: np.ndarray, y: np.ndarray, n_samples: int, random_seed: int) -> RAW_DATASET_TYPE:
    """Data sampling by number of samples

    Args:
        X (np.ndarray): input data
        y (np.ndarray): label of input data
        n_samples (int): number of samples to be extracted
        random_seed (int): random seed

    Returns:
        RAW_DATASET_TYPE: sampled data
    """
    rng = np.random.default_rng(random_seed)
    indices = rng.choice(X.shape[0], n_samples, replace=False)

    return X[indices], y[indices]


def sampling_by_each_class(
    X: np.ndarray, y: np.ndarray, n_samples: int, labels: list[int], random_seed: int
) -> RAW_DATASET_TYPE:
    """Data sampling by each class

    Args:
        X (np.ndarray): input data
        y (np.ndarray): label of input data
        n_samples (int): number of samples to be extracted
        labels (list[int]): labels to be extracted
        random_seed (int): random seed

    Returns:
        RAW_DATASET_TYPE: sampled data
    """
    not_exist_labels = set(labels) - set(map(int, np.unique(y)))
    if not_exist_labels:
        raise ValueError(f"Labels {not_exist_labels} do not exist in the dataset.")

    # fix random seed and shuffle
    rng = np.random.default_rng(random_seed)
    indices = np.arange(X.shape[0])
    rng.shuffle(indices)

    X_shuffled = X[indices]
    y_shuffled = y[indices]

    # label convert to int type
    y_shuffled = np.array([int(label) for label in y_shuffled])
    indices = np.where(np.isin(y_shuffled, labels))[0]
    X_sampled, y_sampled = X_shuffled[indices][:n_samples], y_shuffled[indices][:n_samples]

    return X_sampled, y_sampled



from logging import Logger
from pathlib import Path
from typing import Any, Literal, Optional, cast

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from qxmt.constants import DEFAULT_COLOR_MAP
from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)
RUN_ID_COL = "run_id"


def _check_existence_of_metrics(df: pd.DataFrame, metrics: list[str], logger: Logger = LOGGER) -> list[str]:
    """Check if the target metrics exist in the DataFrame columns.

    Args:
        df (pd.DataFrame): dataframe that contains the calculated metrics
        metrics (list[str]): target metric names
        logger (Logger, optional): logger object. Defaults to LOGGER.

    Returns:
        list[str]: valid metric names
    """
    valid_metrics = []
    for metric in metrics:
        if metric not in df.columns:
            # raise ValueError(f"{metric} is not in the DataFrame columns.")
            logger.warning(f'"{metric}" is not in the DataFrame columns. skip this metric.')
        else:
            valid_metrics.append(metric)

    return valid_metrics


def plot_metric(
    df: pd.DataFrame,
    metric: str,
    chart_type: Literal["bar", "line"] = "bar",
    run_id_col: str = RUN_ID_COL,
    run_ids: Optional[list[int]] = None,
    save_path: Optional[Path] = None,
    logger: Logger = LOGGER,
    **kwargs: Any,
) -> None:
    """Plot bar chart of the target metric. x-axis is run_id.

    Args:
        df (pd.DataFrame): dataframe that contains the calculated metrics
        metric (str): target metric name
        chart_type (Literal["bar", "line"], optional): chart type ("bar" or "line"). Defaults to "bar".
        run_id_col (str, optional): column name of run_id. Defaults to RUN_ID_COL.
        run_ids (Optional[list[int]], optional): run_ids to plot. Defaults to None.
        save_path (Optional[Path], optional): save path for the plot. Defaults to None.
        logger (Logger, optional): logger object. Defaults to LOGGER.
        **kwargs (Any): additional arguments for plot.
            xlabel (Optional[str], optional): x-axis label. Defaults to "run_id".
            ylabel (Optional[str], optional): y-axis label. Defaults to f'"{valid_metric}" score'.
            ylim (Optional[tuple[float, float]], optional): y-axis limit. Defaults to None.
            grid (Optional[bool], optional): grid option. Defaults to True.
            title (Optional[str], optional): title of the plot. Defaults to None.
    """
    if run_ids is not None:
        df = df[df[run_id_col].isin(run_ids)]

    valid_metric = _check_existence_of_metrics(df, [metric], logger)[0]

    plt.figure(figsize=(10, 6), tight_layout=True)
    x = [i for i in range(len(df))]
    color = kwargs.get("color", sns.color_palette(DEFAULT_COLOR_MAP)[0])
    if chart_type == "bar":
        plt.bar(x, df[valid_metric], color=color)
    elif chart_type == "line":
        plt.plot(x, df[valid_metric], color=color, marker="o")
    else:
        raise ValueError(f"chart_type '{chart_type}' is not supported.")
    plt.xlabel(str(kwargs.get("xlabel", run_id_col)))
    plt.ylabel(str(kwargs.get("ylabel", f'"{valid_metric}" score')))
    plt.xticks(x, list(df[run_id_col]))

    ylim = kwargs.get("ylim", (0.0, df[valid_metric].max() * 1.2))
    plt.ylim(*ylim)

    if chart_type == "line":
        grid = kwargs.get("grid", True)
        plt.grid(grid)

    title = cast(str | None, kwargs.get("title", None))
    if title is not None:
        plt.title(str(title))

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()


def plot_metrics_side_by_side(
    df: pd.DataFrame,
    metrics: list[str],
    chart_type: Literal["bar", "line"] = "bar",
    run_id_col: str = RUN_ID_COL,
    run_ids: Optional[list[int]] = None,
    save_path: Optional[Path] = None,
    logger: Logger = LOGGER,
    **kwargs: Any,
) -> None:
    """Plot bar chart of the target metrics side by side. x-axis is run_id.

    Args:
        df (pd.DataFrame): dataframe that contains the calculated metrics
        metrics (list[str]): target metric names
        chart_type (Literal["bar", "line"], optional): chart type ("bar" or "line"). Defaults to "bar".
        run_id_col (str, optional): column name of run_id. Defaults to RUN_ID_COL.
        run_ids (Optional[list[int]], optional): run_ids to plot. Defaults to None.
        save_path (Optional[Path], optional): save path for the plot. Defaults to None.
        logger (Logger, optional): logger object. Defaults to LOGGER.
        **kwargs (Any): additional arguments for plot.
            xlabel (Optional[str], optional): x-axis label. Defaults to "run_id".
            ylabel (Optional[str], optional): y-axis label. Defaults to "metrics score".
            ylim (Optional[tuple[float, float]], optional): y-axis limit. Defaults to None.
            grid (Optional[bool], optional): grid option. Defaults to True.
            title (Optional[str], optional): title of the plot. Defaults to None.
    """
    if run_ids is not None:
        df = df[df[run_id_col].isin(run_ids)]

    valid_metrics = _check_existence_of_metrics(df, metrics, logger)

    plt.figure(figsize=(10, 6), tight_layout=True)
    width = 1.0 / (len(valid_metrics) + 1)
    color: list = cast(list, kwargs.get("color", sns.color_palette(DEFAULT_COLOR_MAP)))
    y_max = 0.0
    for i, metric in enumerate(valid_metrics):
        if chart_type == "bar":
            x = [j + i * width for j in range(len(df))]
            plt.bar(x, df[metric], width=width, label=metric, color=color[i])
        elif chart_type == "line":
            x = [j for j in range(len(df))]
            plt.plot(x, df[metric], label=metric, color=color[i], marker="o")
        else:
            raise ValueError(f"chart_type '{chart_type}' is not supported.")
        y_max = max(y_max, max(df[metric]))

    plt.ylabel(str(kwargs.get("ylabel", "metrics score")))
    plt.xlabel(str(kwargs.get("xlabel", run_id_col)))
    if chart_type == "bar":
        x_ticks = [i + width * 0.5 * (len(valid_metrics) - 1) for i in range(len(df))]
    else:
        x_ticks = [i for i in range(len(df))]
    plt.xticks(x_ticks, list(df[run_id_col]))

    ylim = kwargs.get("ylim", (0.0, y_max * 1.2))
    plt.ylim(*ylim)

    plt.legend(loc="upper right")

    if chart_type == "line":
        grid = kwargs.get("grid", True)
        plt.grid(grid)

    title = cast(str | None, kwargs.get("title", None))
    if title is not None:
        plt.title(str(title))

    if save_path is not None:
        plt.savefig(save_path, bbox_inches="tight", pad_inches=0.1)

    plt.show()



import os

if os.getenv("USE_LLM", "FALSE").lower() == "true":
    from qxmt.generators.description import DescriptionGenerator

    __all__ = [
        "DescriptionGenerator",
    ]
else:
    __all__ = []



from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Optional

import numpy as np

from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class BaseMetric(ABC):
    """Base class for evaluation metric.
    This class is used to define the evaluation metric for the model and visualization.
    Provide a common interface within the QXMT library by absorbing differences between metrics.

    Examples:
        >>> import numpy as np
        >>> from qxmt.evaluation.base import BaseMetric
        >>> class CustomMetric(BaseMetric):
        ...     def __init__(self, name: str) -> None:
        ...         super().__init__(name)
        ...
        ...     @staticmethod
        ...     def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        ...         return np.mean(np.abs(actual - predicted))
        ...
        >>> metric = CustomMetric("mean_absolute_error")
        >>> metric.set_score(np.array([1, 3, 3]), np.array([1, 2, 3]))
        >>> metric.output_score()
        mean_absolute_error: 0.33
    """

    def __init__(self, name: str) -> None:
        """Base class for evaluation metric.

        Args:
            name (str): name of the metric. It is used for the column name of the output and DataFrame.
        """
        self.name: str = name
        self.score: Optional[float] = None

    @staticmethod
    @abstractmethod
    def evaluate(actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> float:
        """define evaluation method for each metric.

        Args:
            actual (np.ndarray): array of actual value
            predicted (np.ndarray): array of predicted value
            **kwargs (dict): additional arguments

        Returns:
            float: evaluated score
        """
        pass

    def set_score(self, actual: np.ndarray, predicted: np.ndarray, **kwargs: Any) -> None:
        """Evaluated the score and set it to the score attribute.

        Args:
            actual (np.ndarray): array of actual value
            predicted (np.ndarray): array of predicted value
            **kwargs (dict): additional arguments
        """
        self.score = self.evaluate(actual, predicted, **kwargs)

    def output_score(self, logger: Logger = LOGGER) -> None:
        """Output the evaluated score on standard output.

        Args:
            logger (Logger, optional): logger object. Defaults to LOGGER.

        Raises:
            ValueError: if the score is not evaluated yet
        """
        if self.score is None:
            raise ValueError("Score is not evaluated yet.")

        logger.info(f"{self.name}: {self.score:.2f}")



import numpy as np
import pennylane as qml

from qxmt.constants import PENNYLANE_PLATFORM
from qxmt.feature_maps.base import BaseFeatureMap


class RotationFeatureMap(BaseFeatureMap):
    """Multi-axis rotation feature map class.

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.rotation import RotationFeatureMap
        >>> feature_map = RotationFeatureMap(2, 2, ["X", "Y"])
        >>> feature_map.draw(x_dim=2)
        0: ─╭AngleEmbedding(M0)─╭AngleEmbedding(M0)─╭AngleEmbedding(M0)─╭AngleEmbedding(M0)─┤
        1: ─╰AngleEmbedding(M0)─╰AngleEmbedding(M0)─╰AngleEmbedding(M0)─╰AngleEmbedding(M0)─┤
        M0 = [0.87336112 0.05976227]
    """

    def __init__(self, n_qubits: int, reps: int, rotation_axis: list[str]) -> None:
        """Initialize the multi axis rotation feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
            rotation_axis (list[str]): list of rotation axis
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.reps: int = reps
        self.rotation_axis: list[str] = rotation_axis

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of rotation feature map.

        Args:
            x (np.ndarray): input data
        """
        for _ in range(self.reps):
            for ax in self.rotation_axis:
                qml.AngleEmbedding(x, wires=range(self.n_qubits), rotation=ax)


class HRotationFeatureMap(BaseFeatureMap):
    """Hadamard and multi-axis rotation feature map class.

    Args:
        BaseFeatureMap (_type_): base feature map class

    Example:
        >>> import numpy as np
        >>> from qxmt.feature_maps.pennylane.rotation import HRotationFeatureMap
        >>> feature_map = HRotationFeatureMap(2, 2, ["X", "Y"])
        >>> feature_map.draw(x_dim=2)
        0: ──H─╭AngleEmbedding(M0)─╭AngleEmbedding(M0)──H─╭AngleEmbedding(M0)─╭AngleEmbedding(M0)─┤
        1: ──H─╰AngleEmbedding(M0)─╰AngleEmbedding(M0)──H─╰AngleEmbedding(M0)─╰AngleEmbedding(M0)─┤
        M0 = [0.88859928 0.32907378]
    """

    def __init__(self, n_qubits: int, reps: int, rotation_axis: list[str]) -> None:
        """Initialize the Hadamard and multi axis rotation feature map class.

        Args:
            n_qubits (int): number of qubits
            reps (int): number of repetitions
            rotation_axis (list[str]): list of rotation axis
        """
        super().__init__(PENNYLANE_PLATFORM, n_qubits)
        self.reps: int = reps
        self.rotation_axis: list[str] = rotation_axis

    def feature_map(self, x: np.ndarray) -> None:
        """Create quantum circuit of Hadamard and rotation feature map.

        Args:
            x (np.ndarray): input data
        """
        for _ in range(self.reps):
            for i in range(self.n_qubits):
                qml.Hadamard(wires=i)
            for ax in self.rotation_axis:
                qml.AngleEmbedding(x, wires=range(self.n_qubits), rotation=ax)



from logging import INFO, Logger, StreamHandler, getLogger


def set_default_logger(logger_name: str) -> Logger:
    logger = getLogger(logger_name)
    logger.setLevel(INFO)
    handler = StreamHandler()
    handler.setLevel(INFO)
    logger.addHandler(handler)

    return logger



from qxmt.feature_maps.pennylane.ising import XXFeatureMap, YYFeatureMap, ZZFeatureMap
from qxmt.feature_maps.pennylane.npqc import NPQCFeatureMap
from qxmt.feature_maps.pennylane.rotation import HRotationFeatureMap, RotationFeatureMap
from qxmt.feature_maps.pennylane.yzcx import YZCXFeatureMap

__all__ = [
    "XXFeatureMap",
    "YYFeatureMap",
    "ZZFeatureMap",
    "HRotationFeatureMap",
    "RotationFeatureMap",
    "YZCXFeatureMap",
    "NPQCFeatureMap",
]



import numpy as np

from qxmt.types import RAW_DATASET_TYPE


def filter_by_categorical(X: np.ndarray, y: np.ndarray, drop_na: bool = True) -> RAW_DATASET_TYPE:
    """Filter the dataset by categorical features.

    Args:
        X (np.ndarray): input data
        y (np.ndarray): label of input data
        drop_na (bool, optional): drop NaN values. Defaults to True.

    Returns:
        RAW_DATASET_TYPE: filtered data
    """
    numerical_indices = []
    for i in range(X.shape[1]):
        try:
            X[:, i].astype(float)
            numerical_indices.append(i)
        except ValueError:
            # if cannot convert to float, it regards as categorical feature
            pass

    X_filtered = X[:, numerical_indices]

    if drop_na:
        X_filtered = np.array(X_filtered, dtype=float)
        mask = ~np.isnan(X_filtered).any(axis=1)
        X_filtered = X_filtered[mask]
        y_filtered = y[mask]

    return X_filtered, y_filtered



from typing import Optional

import numpy as np
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from qxmt.types import PROCESSCED_DATASET_TYPE


def dimension_reduction_by_pca(
    X_train: np.ndarray,
    y_train: np.ndarray,
    X_val: Optional[np.ndarray],
    y_val: Optional[np.ndarray],
    X_test: np.ndarray,
    y_test: np.ndarray,
    n_components: int,
    random_seed: Optional[int] = None,
) -> PROCESSCED_DATASET_TYPE:
    """Dimension reduction by PCA

    Args:
        X_train (np.ndarray): numpy array of training data
        y_train (np.ndarray): numpy array of training label
        X_val (Optional[np.ndarray]): numpy array of validation data. None if validation set is not used
        y_val (Optional[np.ndarray]): numpy array of validation data. None if validation set is not used
        X_test (np.ndarray): numpy array of testing data
        y_test (np.ndarray): numpy array of testing label
        n_components (int): number of components to keep

    Returns:
        PROCESSCED_DATASET_TYPE: tuple of processed dataset
    """
    scaler = StandardScaler()
    scaler.fit(X_train)
    X_train_scaled = scaler.transform(X_train)
    X_val_scaled = scaler.transform(X_val) if X_val is not None else None
    X_test_scaled = scaler.transform(X_test)

    pca = PCA(n_components=n_components, random_state=random_seed)
    pca.fit(X_train_scaled)
    X_train_pca = pca.transform(X_train_scaled)
    X_val_pca = pca.transform(X_val_scaled) if X_val_scaled is not None else None
    X_test_pca = pca.transform(X_test_scaled)

    return X_train_pca, y_train, X_val_pca, y_val, X_test_pca, y_test



import inspect
from logging import Logger
from typing import Any, Callable, Optional

import numpy as np
import optuna
from optuna import Trial
from optuna.samplers import GridSampler, RandomSampler, TPESampler
from sklearn.model_selection import cross_val_score

from qxmt.logger import set_default_logger

LOGGER = set_default_logger(__name__)


class HyperParameterSearch:
    """Hyperparameter search class for machine models using optuna.
    This class provides grid search , random search, and TPE search for hyperparameter optimization.
    The search space is defined by the user, and the search arguments can be customized.
    Reference: https://optuna.readthedocs.io/en/stable/reference/index.html

    Example:
        >>> from sklearn.svm import SVC
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.model_selection import train_test_split
        >>> from qxmt.models.hyperparameter_search.search import HyperParameterSearch
        >>> X, y = load_iris(return_X_y=True)
        >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)
        >>> model = SVC()
        >>> search_type = "tpe"
        >>> search_space = {
        ...     "C": [0.1, 1.0],
        ...     "gamma": [0.01, 0.1]
        ... }
        >>> search_args = {
        ...     "cv": 5,
        ...     "direction": "maximize",
        ...     "n_jobs": -1,
        ... }
        >>> searcher = HyperParameterSearch(X_train, y_train, model, search_type, search_space, search_args)
        >>> best_params = searcher.search()
        >>> best_params
        {'C': 0.8526745595533768, 'gamma': 0.01217052619278743}
    """

    def __init__(
        self,
        X: np.ndarray,
        y: np.ndarray,
        model: Any,
        search_type: str,
        search_space: dict[str, list[Any]],
        search_args: Optional[dict[str, Any]],
        objective: Optional[Callable] = None,
        logger: Logger = LOGGER,
    ) -> None:
        """Initialize the hyperparameter search class.

        Args:
            X (np.ndarray): dataset for search
            y (np.ndarray): target values for search
            model (Any): model instance for hyperparameter search
            search_type (str): search type for hyperparameter search (random, grid, tpe)
            search_space (dict[str, list[Any]]): search space for hyperparameter search
            search_args (Optional[dict[str, Any]]): search arguments for hyperparameter search
            objective (Optional[Callable], optional): objective function for hyperparameter search. Defaults to None.
            logger (Logger, optional): logger instance. Defaults to LOGGER.

        Raises:
            ValueError: Sampler of search_args not matching search type
        """
        self.X = X
        self.y = y
        self.model = model
        self.search_type = search_type
        self.search_space = search_space
        self.search_args = search_args or {}
        self.objective = objective
        self.logger = logger

        self.scoring = self.search_args.get("scoring")

        sampler_from_type = self._get_sampler()
        if (search_args is not None) and (
            not isinstance(search_args.get("sampler", sampler_from_type), type(sampler_from_type))
        ):
            raise ValueError(
                f"""Sampler '{search_args['sampler']}' specified in search_args is
                 not matching search_type '{self.search_type}'."""
            )

    def _get_sampler(self) -> RandomSampler | GridSampler | TPESampler:
        """Get sampler instance based on search_type. If not specified, default to TPE sampler.

        Returns:
            Any: sampler instance
        """
        match self.search_type.lower():
            case "random":
                return RandomSampler()
            case "grid":
                return GridSampler(self.search_space)
            case "tpe":
                return TPESampler()
            case _:
                self.logger.warning(f"Unknown search_type '{self.search_type}', defaulting to TPE sampler.")
                return TPESampler()

    def _get_direction(self) -> str:
        """Get direction based on scoring. If not specified, default to 'maximize'.

        Returns:
            str: direction of optimization
        """
        direction = self.search_args.get("direction", None)
        if direction is None:
            if self.scoring in ["neg_mean_squared_error", "neg_log_loss"]:
                direction = "minimize"
            else:
                direction = "maximize"
            self.logger.info(f"Direction not specified, defaulting to '{direction}' based on scoring '{self.scoring}'.")

        return direction

    def _get_study_args(self) -> dict[str, Any]:
        """Get study arguments for optuna study for search_args.

        Returns:
            dict[str, Any]: optuna study arguments
        """
        study_args = {}
        if self.search_args is not None:
            study_args = {
                key: value
                for key, value in self.search_args.items()
                if key in inspect.signature(optuna.create_study).parameters.keys()
            }

        if study_args.get("sampler") is None:
            study_args["sampler"] = self._get_sampler()

        if study_args.get("direction") is None:
            study_args["direction"] = self._get_direction()

        return study_args

    def _get_optimize_args(self) -> dict[str, Any]:
        """Get optimize arguments for optuna study for search_args.

        Returns:
            dict[str, Any]: optuna optimize arguments
        """
        optimize_args = {}
        if self.search_args is not None:
            optimize_args = {
                key: value
                for key, value in self.search_args.items()
                if key in inspect.signature(optuna.study.Study.optimize).parameters.keys()
            }
        return optimize_args

    def search(self) -> dict[str, Any]:
        """Search the best hyperparameters for the model.

        Returns:
            dict[str, Any]: best hyperparameters found by search
        """
        study_args = self._get_study_args()
        study = optuna.create_study(**study_args)

        if self.objective is None:
            self.objective = self.default_objective

        optimize_args = self._get_optimize_args()
        study.optimize(self.objective, **optimize_args)

        self.logger.info(f"Best parameters: {study.best_params}")
        self.logger.info(f"Best value: {study.best_value}")

        return study.best_params

    def _format_params_to_optuna(self, trial: Trial) -> dict[str, Any]:
        """Set the parameters in the format of optuna.

        Args:
            trial (Trial): optuna trial instance

        Returns:
            dict[str, Any]: parameters in the format of optuna
        """
        params = {}
        for param_name, param_values in self.search_space.items():
            if len(param_values) == 2 and all(isinstance(v, int) for v in param_values):
                # define range of integers
                low, high = param_values
                params[param_name] = trial.suggest_int(param_name, low, high)
            elif len(param_values) == 2 and all(isinstance(v, float) for v in param_values):
                # define range of floats
                low, high = param_values
                params[param_name] = trial.suggest_float(param_name, low, high)
            else:
                # define categorical values
                params[param_name] = trial.suggest_categorical(param_name, param_values)

        return params

    def default_objective(self, trial: Trial) -> float:
        """Default objective function for hyperparameter search.

        Args:
            trial (Trial): optuna trial instance

        Returns:
            float: score of the model
        """
        params = self._format_params_to_optuna(trial)
        model = self.model.set_params(**params)

        # evaluate the model by mean score of cross validation
        cv = self.search_args.get("cv", 3)
        score = cross_val_score(model, self.X, self.y, cv=cv, scoring=self.scoring)

        return score.mean()
